import itertools
import warnings

import biotite.structure as struc
import biotite.structure.io.pdbx as pdbx
import numpy as np
from biotite import InvalidFileError

from ..molecule.assembly import AssemblyParser
from ..molecule.molecule import Molecule


class OldCIF(Molecule):
    def __init__(self, file_path, extra_fields=None, sec_struct=True):
        super().__init__(file_path=file_path)
        self.array = self._get_structure(
            extra_fields=extra_fields, sec_struct=sec_struct
        )
        self.n_atoms = self.array.array_length()

    def _read(self, file_path):
        return pdbx.legacy.PDBxFile.read(file_path)

    def _get_structure(self, extra_fields: str = None, sec_struct=True, bonds=True):
        fields = ["b_factor", "charge", "occupancy", "atom_id"]
        if extra_fields:
            [fields.append(x) for x in extra_fields]

        # if the 'atom_site' doesn't exist then it will just be a small molecule
        # which can be extracted with the get_component()
        try:
            array = pdbx.get_structure(self.file, extra_fields=extra_fields)
            annotations = {
                "sec_struct": _get_secondary_structure,
                "entity_id": _get_entity_id,
            }
            for key, func in annotations.items():
                try:
                    array.set_annotation(key, func(array, self.file))
                except KeyError as e:
                    print(e)

        except InvalidFileError:
            array = pdbx.get_component(self.file)

        # pdbx files don't seem to have bond information defined, so connect them based
        # on their residue names
        if not array.bonds and bonds:
            array.bonds = struc.bonds.connect_via_residue_names(
                array, inter_residue=True
            )

        return array

    def _entity_ids(self):
        entities = self.file["entity"]
        if not entities:
            return None

        return entities.get("pdbx_description", None)

    def _assemblies(self):
        return CIFAssemblyParser(self.file).get_assemblies()


def _ss_label_to_int(label):
    if "HELX" in label:
        return 1
    elif "STRN" in label:
        return 2
    else:
        return 3


def _get_secondary_structure(array, file):
    """
    Get secondary structure information for the array from the file.

    Parameters
    ----------
    array : numpy array
        The array for which secondary structure information is to be retrieved.
    file : object
        The file object containing the secondary structure information.

    Returns
    -------
    numpy array
        A numpy array of secondary structure information, where each element is either 0, 1, 2, or 3.
        - 0: Not a peptide
        - 1: Alpha helix
        - 2: Beta sheet
        - 3: Loop

    Raises
    ------
    KeyError
        If the 'struct_conf' category is not found in the file.
    """

    # get the annotations for the struc_conf cetegory. Provides start and end
    # residues for the annotations. For most files this will only contain the
    # alpha helices, but will sometimes contain also other secondary structure
    # information such as in AlphaFold predictions

    conf = file.get_category("struct_conf")
    if not conf:
        raise KeyError
    starts = conf["beg_auth_seq_id"].astype(int)
    ends = conf["end_auth_seq_id"].astype(int)
    chains = conf["end_auth_asym_id"].astype(str)
    id_label = conf["id"].astype(str)

    # most files will have a separate category for the beta sheets
    # this can just be appended to the other start / end / id and be processed
    # as normal
    sheet = file.get_category("struct_sheet_range")
    if sheet:
        starts = np.append(starts, sheet["beg_auth_seq_id"].astype(int))
        ends = np.append(ends, sheet["end_auth_seq_id"].astype(int))
        chains = np.append(chains, sheet["end_auth_asym_id"].astype(str))
        id_label = np.append(id_label, np.repeat("STRN", len(sheet["id"])))

    # convert the string labels to integer representations of the SS
    # AH: 1, BS: 2, LOOP: 3

    id_int = np.array([_ss_label_to_int(label) for label in id_label], int)

    # create a lookup dictionary that enables lookup of secondary structure
    # based on the chain_id and res_id values

    lookup = dict()
    for chain in np.unique(chains):
        arrays = []
        mask = chain == chains
        start_sub = starts[mask]
        end_sub = ends[mask]
        id_sub = id_int[mask]

        for start, end, id in zip(start_sub, end_sub, id_sub):
            idx = np.arange(start, end + 1, dtype=int)
            arr = np.zeros((len(idx), 2), dtype=int)
            arr[:, 0] = idx
            arr[:, 1] = 3
            arr[:, 1] = id
            arrays.append(arr)

        lookup[chain] = dict(np.vstack(arrays).tolist())

    # use the lookup dictionary to get the SS annotation based on the chain_id and res_id
    secondary_structure = np.zeros(len(array.chain_id), int)
    for i, (chain, res) in enumerate(zip(array.chain_id, array.res_id)):
        try:
            secondary_structure[i] = lookup[chain].get(res, 3)
        except KeyError:
            secondary_structure[i] = 0

    # assign SS to 0 where not peptide
    secondary_structure[~struc.filter_amino_acids(array)] = 0
    return secondary_structure


def _get_entity_id(array, file):
    entities = file.get_category("entity_poly")
    if not entities:
        raise KeyError
    chain_ids = entities["pdbx_strand_id"]

    # the chain_ids are an array of individual items np.array(['A,B', 'C', 'D,E,F'])
    # which need to be categorised as [1, 1, 2, 3, 3, 3] for their belonging to individual
    # entities

    chains = []
    idx = []
    for i, chain_str in enumerate(chain_ids):
        for chain in chain_str.split(","):
            chains.append(chain)
            idx.append(i)

    entity_lookup = dict(zip(chains, idx))
    chain_id_int = np.array(
        [entity_lookup.get(chain, -1) for chain in array.chain_id], int
    )
    return chain_id_int


class CIFAssemblyParser(AssemblyParser):
    # Implementation adapted from ``biotite.structure.io.pdbx.convert``

    def __init__(self, file_cif):
        self._file = file_cif

    def list_assemblies(self):
        return list(pdbx.list_assemblies(self._file).keys())

    def get_transformations(self, assembly_id):
        assembly_gen_category = self._file["pdbx_struct_assembly_gen"]

        struct_oper_category = self._file["pdbx_struct_oper_list"]

        if assembly_id not in assembly_gen_category["assembly_id"]:
            raise KeyError(f"File has no Assembly ID '{assembly_id}'")

        # Extract all possible transformations indexed by operation ID
        transformation_dict = _get_transformations(struct_oper_category)

        # Get necessary transformations and the affected chain IDs
        # NOTE: The chains given here refer to the `label_asym_id` field
        # of the `atom_site` category
        # However, by default `PDBxFile` uses the `auth_asym_id` as
        # chain ID
        matrices = []
        for id, op_expr, asym_id_expr in zip(
            assembly_gen_category["assembly_id"],
            assembly_gen_category["oper_expression"],
            assembly_gen_category["asym_id_list"],
        ):
            # Find the operation expressions for given assembly ID
            # We already asserted that the ID is actually present
            if id == assembly_id:
                operations = _parse_operation_expression(op_expr)
                affected_chain_ids = asym_id_expr.split(",")
                for i, operation in enumerate(operations):
                    rotations = []
                    translations = []
                    for op_step in operation:
                        rotation, translation = transformation_dict[op_step]
                        rotations.append(rotation)
                        translations.append(translation)
                    matrix = _chain_transformations(rotations, translations)
                    matrices.append((affected_chain_ids, matrix.tolist()))

        return matrices

    def get_assemblies(self):
        assembly_dict = {}
        for assembly_id in self.list_assemblies():
            assembly_dict[assembly_id] = self.get_transformations(assembly_id)

        return assembly_dict


def _chain_transformations(rotations, translations):
    """
    Get a total rotation/translation transformation by combining
    multiple rotation/translation transformations.
    This is done by intermediately combining rotation matrices and
    translation vectors into 4x4 matrices in the form

    |r11 r12 r13 t1|
    |r21 r22 r23 t2|
    |r31 r32 r33 t3|
    |0   0   0   1 |.
    """
    total_matrix = np.identity(4)
    for rotation, translation in zip(rotations, translations):
        matrix = np.zeros((4, 4))
        matrix[:3, :3] = rotation
        matrix[:3, 3] = translation
        matrix[3, 3] = 1
        total_matrix = matrix @ total_matrix

    # return total_matrix[:3, :3], total_matrix[:3, 3]
    return matrix


def _get_transformations(struct_oper):
    """
    Get transformation operation in terms of rotation matrix and
    translation for each operation ID in ``pdbx_struct_oper_list``.
    """
    transformation_dict = {}
    for index, id in enumerate(struct_oper["id"]):
        rotation_matrix = np.array(
            [
                [float(struct_oper[f"matrix[{i}][{j}]"][index]) for j in (1, 2, 3)]
                for i in (1, 2, 3)
            ]
        )
        translation_vector = np.array(
            [float(struct_oper[f"vector[{i}]"][index]) for i in (1, 2, 3)]
        )
        transformation_dict[id] = (rotation_matrix, translation_vector)
    return transformation_dict


def _parse_operation_expression(expression):
    """
    Get successive operation steps (IDs) for the given
    ``oper_expression``.
    Form the cartesian product, if necessary.
    """
    # Split groups by parentheses:
    # use the opening parenthesis as delimiter
    # and just remove the closing parenthesis
    expressions_per_step = expression.replace(")", "").split("(")
    expressions_per_step = [e for e in expressions_per_step if len(e) > 0]
    # Important: Operations are applied from right to left
    expressions_per_step.reverse()

    operations = []
    for expr in expressions_per_step:
        if "-" in expr:
            if "," in expr:
                for gexpr in expr.split(","):
                    if "-" in gexpr:
                        first, last = gexpr.split("-")
                        operations.append(
                            [str(id) for id in range(int(first), int(last) + 1)]
                        )
                    else:
                        operations.append([gexpr])
            else:
                # Range of operation IDs, they must be integers
                first, last = expr.split("-")
                operations.append([str(id) for id in range(int(first), int(last) + 1)])
        elif "," in expr:
            # List of operation IDs
            operations.append(expr.split(","))
        else:
            # Single operation ID
            operations.append([expr])

    # Cartesian product of operations
    return list(itertools.product(*operations))



# import bpy
import numpy as np
import pytest
import bpy
import molecularnodes as mn
from molecularnodes.blender import nodes
import random

from .utils import NumpySnapshotExtension
from .constants import codes, data_dir

random.seed(6)


def test_get_nodes():
    mol = mn.entities.fetch("4ozs", style="spheres", cache_dir=data_dir)

    assert nodes.get_nodes_last_output(mol.node_group)[0].name == "Style Spheres"
    nodes.realize_instances(mol.object)
    assert nodes.get_nodes_last_output(mol.node_group)[0].name == "Realize Instances"
    assert nodes.get_style_node(mol.object).name == "Style Spheres"

    mol2 = mn.entities.fetch(
        "1cd3", style="cartoon", build_assembly=True, cache_dir=data_dir
    )

    assert nodes.get_nodes_last_output(mol2.node_group)[0].name == "Assembly 1cd3"
    assert nodes.get_style_node(mol2.object).name == "Style Cartoon"


def test_selection():
    chain_ids = [let for let in "ABCDEFG123456"]
    node = nodes.custom_iswitch(
        "test_node", chain_ids, prefix="Chain ", dtype="BOOLEAN"
    )

    input_sockets = nodes.inputs(node)
    for letter, socket in zip(chain_ids, input_sockets.values()):
        assert f"Chain {letter}" == socket.name
        assert socket.default_value is False


@pytest.mark.parametrize("code", codes)
@pytest.mark.parametrize("attribute", ["chain_id", "entity_id"])
def test_selection_working(snapshot_custom: NumpySnapshotExtension, attribute, code):
    mol = mn.entities.fetch(code, style="ribbon", cache_dir=data_dir)
    group = mol.node_group
    node_sel = nodes.add_selection(
        group, mol.name, mol.object[f"{attribute}s"], attribute
    )

    n = len(node_sel.inputs)

    nodes.realize_instances(mol.object)

    for inp in node_sel.inputs:
        inp.default_value = True
        pos = mol.named_attribute("position", evaluate=True)
        assert snapshot_custom == pos.shape
        assert snapshot_custom == pos
        inp.default_value = False


@pytest.mark.parametrize("code", codes)
@pytest.mark.parametrize("attribute", ["chain_id", "entity_id"])
def test_color_custom(snapshot_custom: NumpySnapshotExtension, code, attribute):
    mol = mn.entities.fetch(code, style="ribbon", cache_dir=data_dir)

    group_col = mn.blender.nodes.custom_iswitch(
        name=f"Color Entity {mol.name}",
        iter_list=mol.object[f"{attribute}s"],
        field=attribute,
        dtype="RGBA",
    )
    group = mol.node_group
    node_col = mn.blender.nodes.add_custom(group, group_col.name, [0, -200])
    group.links.new(node_col.outputs[0], group.nodes["Set Color"].inputs["Color"])
    for i, input in enumerate(node_col.inputs):
        input.default_value = mn.color.random_rgb(i)

    assert snapshot_custom == mol.named_attribute("Color")


def test_custom_resid_selection():
    node = mn.blender.nodes.resid_multiple_selection("new_node", "1, 5, 10-20, 40-100")
    numbers = [1, 5, 10, 20, 40, 100]
    assert len(nodes.outputs(node)) == 2
    counter = 0
    for item in node.interface.items_tree:
        if item.in_out == "INPUT":
            assert item.default_value == numbers[counter]
            counter += 1


def test_iswitch_creation():
    items = list(range(10))
    tree_boolean = mn.blender.nodes.custom_iswitch(
        name="newboolean", iter_list=items, dtype="BOOLEAN"
    )
    # ensure there isn't an item called 'Color' in the created interface
    assert not tree_boolean.interface.items_tree.get("Color")
    assert tree_boolean.interface.items_tree["Selection"].in_out == "OUTPUT"
    assert tree_boolean.interface.items_tree["Inverted"].in_out == "OUTPUT"
    for i in items:
        assert tree_boolean.interface.items_tree[str(i)].in_out == "INPUT"

    tree_rgba = mn.blender.nodes.custom_iswitch(
        name="newcolor", iter_list=items, dtype="RGBA"
    )
    # ensure there isn't an item called 'selection'
    assert not tree_rgba.interface.items_tree.get("Selection")
    assert tree_rgba.interface.items_tree["Color"].in_out == "OUTPUT"
    for i in items:
        assert tree_rgba.interface.items_tree[str(i)].in_out == "INPUT"


def test_op_custom_color():
    mol = mn.entities.load_local(data_dir / "1cd3.cif").object
    mol.select_set(True)
    group = mn.blender.nodes.custom_iswitch(
        name=f"Color Chain {mol.name}", iter_list=mol["chain_ids"], dtype="RGBA"
    )

    assert group
    assert group.interface.items_tree["G"].name == "G"
    assert group.interface.items_tree[-1].name == "G"
    assert group.interface.items_tree[0].name == "Color"


def test_color_lookup_supplied():
    col = mn.color.random_rgb(6)
    name = "test"
    node = mn.blender.nodes.custom_iswitch(
        name=name,
        iter_list=range(10, 20),
        dtype="RGBA",
        default_values=[col for i in range(10)],
        start=10,
    )
    assert node.name == name
    for item in nodes.inputs(node).values():
        assert np.allclose(np.array(item.default_value), col)

    node = mn.blender.nodes.custom_iswitch(
        name="test2", iter_list=range(10, 20), dtype="RGBA", start=10
    )
    for item in nodes.inputs(node).values():
        assert not np.allclose(np.array(item.default_value), col)


def get_links(sockets):
    links = []
    for socket in sockets:
        for link in socket.links:
            yield link


def test_change_style():
    model = mn.entities.fetch("1cd3", style="cartoon", cache_dir=data_dir).object
    style_node_1 = nodes.get_style_node(model).name
    mn.blender.nodes.change_style_node(model, "ribbon")
    style_node_2 = nodes.get_style_node(model).name

    assert style_node_1 != style_node_2

    styles_to_check = ["ribbon", "cartoon", "ball_and_stick", "surface"] + list(
        [f"preset_{i}" for i in [1, 2, 3, 4]]
    )

    for style in styles_to_check:
        style_node_1 = nodes.get_style_node(model)
        links_in_1 = [link.from_socket.name for link in get_links(style_node_1.inputs)]
        links_out_1 = [
            link.from_socket.name for link in get_links(style_node_1.outputs)
        ]

        nodes.change_style_node(model, style)
        style_node_2 = nodes.get_style_node(model)
        links_in_2 = [link.from_socket.name for link in get_links(style_node_2.inputs)]
        links_out_2 = [
            link.from_socket.name for link in get_links(style_node_2.outputs)
        ]

        assert len(links_in_1) == len(links_in_2)
        assert len(links_out_1) == len(links_out_2)


@pytest.fixture
def pdb_8h1b():
    return mn.entities.fetch("8H1B", del_solvent=False, cache_dir=data_dir, style=None)


topology_node_names = mn.ui.node_info.menu_items.get_submenu("topology").node_names()


def test_nodes_exist():
    for menu in mn.ui.node_info.menu_items.submenus:
        for item in menu.items:
            if item.is_break or item.is_custom:
                continue
            if item.name.startswith("mn."):
                continue
            mn.blender.nodes.append(item.name)
            assert True


@pytest.mark.parametrize("node_name", topology_node_names)
@pytest.mark.parametrize("code", codes)
def test_node_topology(snapshot_custom: NumpySnapshotExtension, code, node_name):
    mol = mn.entities.fetch(code, del_solvent=False, cache_dir=data_dir, style=None)

    group = nodes.get_mod(mol.object).node_group = nodes.new_tree()

    group.links.new(
        group.nodes["Group Input"].outputs[0], group.nodes["Group Output"].inputs[0]
    )
    node_att = group.nodes.new("GeometryNodeStoreNamedAttribute")
    node_att.inputs[2].default_value = "test_attribute"
    nodes.insert_last_node(group, node_att)
    # exclude these particular nodes, as they aren't field nodes and so we shouldn't
    # be testing them here. Will create their own particular tests later
    if any(
        keyword in node_name
        for keyword in ["Bonds", "Bond Count", "DSSP", "Chain Group ID"]
    ):
        return None

    node_topo = nodes.add_custom(
        group, node_name, location=[x - 300 for x in node_att.location]
    )

    if node_name == "Residue Mask":
        node_topo.inputs["atom_name"].default_value = 61

    type_to_data_type = {
        "VECTOR": "FLOAT_VECTOR",
        "VALUE": "FLOAT",
        "BOOLEAN": "BOOLEAN",
        "INT": "INT",
        "RGBA": "FLOAT_COLOR",
        "ROTATION": "QUATERNION",
    }

    for output in node_topo.outputs:
        node_att.data_type = type_to_data_type[output.type]
        input = node_att.inputs["Value"]

        for link in input.links:
            group.links.remove(link)

        group.links.new(output, input)

        assert snapshot_custom == mol.named_attribute("test_attribute", evaluate=True)


def test_topo_bonds():
    mol = mn.entities.fetch("1BNA", del_solvent=True, style=None, cache_dir=data_dir)
    group = nodes.get_mod(mol.object).node_group = nodes.new_tree()

    # add the node that will break bonds, set the cutoff to 0
    node_break = nodes.add_custom(group, "Topology Break Bonds")
    nodes.insert_last_node(group, node=node_break)
    node_break.inputs["Cutoff"].default_value = 0

    # compare the number of edges before and after deleting them with
    bonds = mol.object.data.edges
    no_bonds = mol.evaluate().data.edges
    assert len(bonds) > len(no_bonds)
    assert len(no_bonds) == 0

    # add the node to find the bonds, and ensure the number of bonds pre and post the nodes
    # are the same (other attributes will be different, but for now this is good)
    node_find = nodes.add_custom(group, "Topology Find Bonds")
    nodes.insert_last_node(group, node=node_find)
    bonds_new = mol.evaluate().data.edges
    assert len(bonds) == len(bonds_new)


def test_is_modifier():
    bpy.ops.wm.open_mainfile(filepath=mn.utils.MN_DATA_FILE)
    for tree in bpy.data.node_groups:
        if hasattr(tree, "is_modifier"):
            assert not tree.is_modifier
    mol = mn.entities.fetch("4ozs")
    assert mol.tree.is_modifier



import bpy
from bpy.types import PropertyGroup
from bpy.props import IntProperty, BoolProperty, EnumProperty, StringProperty
from .handlers import _update_trajectories
from .style import STYLE_ITEMS


bpy.types.Scene.MN_import_centre = BoolProperty(
    name="Centre Structure",
    description="Move the imported Molecule on the World Origin",
    default=False,
)

bpy.types.Scene.MN_centre_type = EnumProperty(
    name="Method",
    default="mass",
    items=(
        (
            "mass",
            "Mass",
            "Adjust the structure's centre of mass to be at the world origin",
            1,
        ),
        (
            "centroid",
            "Centroid",
            "Adjust the structure's centroid (centre of geometry) to be at the world origin",
            2,
        ),
    ),
)

bpy.types.Scene.MN_import_del_solvent = BoolProperty(
    name="Remove Solvent",
    description="Delete the solvent from the structure on import",
    default=True,
)
bpy.types.Scene.MN_import_panel_selection = IntProperty(
    name="MN_import_panel_selection",
    description="Import Panel Selection",
    subtype="NONE",
    default=0,
)
bpy.types.Scene.MN_import_build_assembly = BoolProperty(
    name="Build Assembly", default=False
)
bpy.types.Scene.MN_import_node_setup = BoolProperty(
    name="Setup Nodes",
    default=True,
    description="Create and set up a Geometry Nodes tree on import",
)

bpy.types.Object.uuid = StringProperty(  # type: ignore
    name="UUID",
    description="Unique ID for referencing the required objects in the MNSession",
    default="",
)


class MolecularNodesSceneProperties(PropertyGroup):
    import_oxdna_topology: StringProperty(  # type: ignore
        name="Toplogy",
        description="File path for the topology to import (.top)",
        subtype="FILE_PATH",
    )
    import_oxdna_trajectory: StringProperty(  # type: ignore
        name="Trajectory",
        description="File path for the trajectory to import (.oxdna / .dat)",
        subtype="FILE_PATH",
    )
    import_oxdna_name: StringProperty(  # type: ignore
        name="Name", description="Name of the created object.", default="NewOrigami"
    )
    import_style: EnumProperty(  # type: ignore
        name="Style",
        description="Default style for importing",
        items=STYLE_ITEMS,
        default="spheres",
    )


class MolecularNodesObjectProperties(PropertyGroup):
    entity_type: EnumProperty(  # type: ignore
        name="Entity Type",
        description="How the file was imported, dictating how MN interacts with it",
        items=(
            ("molecule", "Molecule", "A single molecule"),
            ("ensemble", "Ensemble", "A collection of molecules"),
            ("density", "Density", "An electron density map"),
            ("md", "Trajectory", "A molecular dynamics trajectory"),
            ("md-oxdna", "oxDNA Trajectory", "A oxDNA molecular dynamics trajectory "),
        ),
    )

    pdb_code: StringProperty(  # type: ignore
        name="PDB",
        description="PDB code used to download this structure",
        maxlen=4,
        options={"HIDDEN"},
    )
    trajectory_selection_index: IntProperty(  # type: ignore
        name="Index of selection",
        description="Index of selection, that is selected for the UI",
        default=0,
    )
    frame_hidden: IntProperty(  # type: ignore
        name="Frame",
        description="Frame of the loaded trajectory",
        default=0,
        min=0,
    )
    frame: IntProperty(  # type: ignore
        name="Frame",
        description="Frame of the loaded trajectory",
        default=0,
        update=_update_trajectories,
        min=0,
    )
    update_with_scene: BoolProperty(  # type: ignore
        name="Update with Scene",
        description="Update the trajectory with the scene frame",
        default=True,
        update=_update_trajectories,
    )
    subframes: IntProperty(  # type: ignore
        name="Subframes",
        description="Number of subframes to insert between frames of the loaded trajectory",
        default=0,
        update=_update_trajectories,
        min=0,
    )
    offset: IntProperty(  # type: ignore
        name="Offset",
        description="Offset the starting playback for the trajectory on the timeine. Positive starts the playback later than frame 0, negative starts it earlier than frame 0",
        default=0,
        update=_update_trajectories,
    )
    interpolate: BoolProperty(  # type: ignore
        name="Interpolate",
        description="Whether to interpolate when using subframes",
        default=True,
        update=_update_trajectories,
    )
    average: IntProperty(  # type: ignore
        name="Average",
        description="Average the position this number of frames either side of the current frame",
        default=0,
        update=_update_trajectories,
        min=0,
        soft_max=5,
    )
    correct_periodic: BoolProperty(  # type: ignore
        name="Correct",
        description="Correct for periodic boundary crossing when using interpolation or averaging. Assumes cubic dimensions and only works if the unit cell is orthorhombic",
        default=False,
        update=_update_trajectories,
    )
    filepath_trajectory: StringProperty(  # type: ignore
        name="Trajectory",
        description="Filepath for the `trajectory` of the Object",
        subtype="FILE_PATH",
        default="",
    )
    filepath_topology: StringProperty(  # type: ignore
        name="Topology",
        description="Filepath for the Topology of the Object",
        subtype="FILE_PATH",
        default="",
    )


CLASSES = [MolecularNodesObjectProperties, MolecularNodesSceneProperties]



from typing import Dict, Callable

import bpy
import MDAnalysis as mda
import numpy as np
import numpy.typing as npt
from math import floor, remainder

from enum import Enum

from ... import data
from ..entity import MolecularEntity, EntityType
from ...blender import coll, nodes, path_resolve
import databpy
from ...utils import (
    correct_periodic_positions,
    frame_mapper,
    frames_to_average,
    fraction,
)
from .selections import Selection


class Trajectory(MolecularEntity):
    def __init__(self, universe: mda.Universe, world_scale: float = 0.01):
        super().__init__()
        self.universe: mda.Universe = universe
        self.selections: Dict[str, Selection] = {}
        self.calculations: Dict[str, Callable] = {}
        self.world_scale = world_scale
        self.frame_mapping: npt.NDArray[np.int64] | None = None
        self.cache: dict = {}
        self._entity_type = EntityType.MD

    def selection_from_ui(self, ui_item) -> Selection:
        self.selections[ui_item.name] = Selection(
            universe=self.universe,
            selection_str=ui_item.selection_str,
            name=ui_item.name,
            updating=ui_item.updating,
            periodic=ui_item.periodic,
        )
        self.apply_selection(self.selections[ui_item.name])

        return self.selections[ui_item.name]

    def add_selection(
        self,
        selection_str: str,
        name: str,
        updating: bool = True,
        periodic: bool = True,
    ):
        "Adds a new selection with the given name, selection string and selection parameters."
        obj = self.object

        obj.mn_trajectory_selections.add()
        sel = obj.mn_trajectory_selections[-1]
        sel.name = name
        sel.selection_str = selection_str
        sel.updating = updating
        sel.periodic = periodic

        self.selection_from_ui(sel)

        return sel

    def add_selection_from_atomgroup(self, atomgroup: mda.AtomGroup, name: str = ""):
        "Create a Selection object from an AtomGroup"
        selection = Selection.from_atomgroup(atomgroup, name=name)

        obj = self.object
        obj.mn_trajectory_selections.add()
        sel = obj.mn_trajectory_selections[-1]

        if not atomgroup.__class__.__name__ == "UpdatingAtomGroup":
            sel.immutable = True
        sel.name = selection.name
        sel.selection_str = selection.selection_str
        sel.updating = selection.updating
        sel.periodic = selection.periodic

        self.selections[selection.name] = selection
        self.apply_selection(selection)
        return sel

    def apply_selection(self, selection: Selection):
        "Set the boolean attribute for this selection on the mesh of the object"
        self.set_boolean(selection.to_mask(), name=selection.name)

    @property
    def is_orthorhombic(self):
        dim = self.universe.dimensions
        if dim is None:
            return False

        return np.allclose(dim[3:], 90.0)

    @property
    def atoms(self) -> mda.AtomGroup:
        return self.universe.atoms

    @property
    def n_atoms(self) -> int:
        return self.atoms.n_atoms

    @staticmethod
    def bool_selection(ag, selection, **kwargs) -> np.ndarray:
        return np.isin(ag.ix, ag.select_atoms(selection, **kwargs).ix).astype(bool)

    @property
    def univ_positions(self) -> np.ndarray:
        return self.atoms.positions * self.world_scale

    @property
    def bonds(self) -> np.ndarray:
        # the code to remap indices for a selection was removed as we don't subset the trajectory anymore
        # when importing it, everything is imported and the selections just update
        if hasattr(self.atoms, "bonds"):
            return self.atoms.bonds.indices
        else:
            return None

    @property
    def elements(self) -> np.ndarray:
        if hasattr(self.atoms, "elements"):
            return self.atoms.elements

        try:
            guessed_elements = [
                x
                if x in data.elements.keys()
                else mda.topology.guessers.guess_atom_element(x)
                for x in self.atoms.names
            ]
            return np.array(guessed_elements)

        except Exception:
            return np.repeat("X", self.n_atoms)

    @property
    def atomic_number(self) -> np.ndarray:
        return np.array(
            [
                data.elements.get(element, data.elements.get("X")).get("atomic_number")
                for element in self.elements
            ]
        )

    @property
    def vdw_radii(self) -> np.ndarray:
        return (
            np.array(
                [
                    data.elements.get(element, {}).get("vdw_radii", 100)
                    for element in self.elements
                ]
            )
            * 0.01  # pm to Angstrom
            * self.world_scale  # Angstrom to world scale
        )

    @property
    def mass(self) -> np.ndarray:
        # units: daltons
        if hasattr(self.atoms, "masses"):
            return np.array([x.mass for x in self.atoms])
        else:
            masses = [
                data.elements.get(element, {"standard_mass": 0}).get("standard_mass")
                for element in self.elements
            ]
            return np.array(masses)

    @property
    def n_frames(self) -> int:
        return self.universe.trajectory.n_frames

    @property
    def res_id(self) -> np.ndarray:
        return self.atoms.resids

    @property
    def uframe(self) -> int:
        """
        Get the current frame number of the linked `Universe.trajectory`.

        Returns:
            int: Current frame number in the trajectory.
        """
        return self.universe.trajectory.frame

    @uframe.setter
    def uframe(self, value) -> None:
        """
        Set the current frame number of the linked `Universe.trajectory`.

        The frame number is clamped between 0 and n_frames-1 to prevent
        out-of-bounds access.

        Args:
            value (int): Target frame number to set.

        Returns:
            None
        """
        if self.universe.trajectory.frame != value:
            self.universe.trajectory[value]

    @property
    def res_name(self) -> np.ndarray:
        return np.array(list(map(lambda x: x[0:3], self.atoms.resnames)))

    @property
    def atom_id(self) -> np.ndarray:
        return self.universe.atoms.atom_id

    @property
    def res_num(self) -> np.ndarray:
        return np.array(
            [
                data.residues.get(res_name, data.residues.get("UNK")).get(
                    "res_name_num"
                )
                for res_name in self.res_name
            ]
        )

    @property
    def b_factor(self) -> np.ndarray:
        if hasattr(self.atoms, "tempfactors"):
            return self.atoms.tempfactors
        else:
            return np.zeros(self.n_atoms)

    @property
    def segindices(self) -> np.ndarray:
        if hasattr(self.atoms, "segindices"):
            return self.atoms.segindices

    @property
    def chain_id(self) -> np.ndarray:
        if hasattr(self.atoms, "chainIDs"):
            return self.atoms.chainIDs
        else:
            return np.zeros(self.n_atoms)

    @property
    def chain_ids(self) -> np.ndarray:
        return np.unique(self.chain_id)

    @property
    def chain_id_num(self) -> np.ndarray:
        chain_ids, chain_id_index = np.unique(self.chain_id, return_inverse=True)
        return chain_id_index

    @property
    def atom_type(self) -> np.ndarray:
        return self.atoms.types

    @property
    def atom_type_unique(self) -> np.ndarray:
        return np.unique(self.atom_type)

    @property
    def atom_type_num(self) -> np.ndarray:
        try:
            atom_type_unique, atom_type_index = np.unique(
                self.atom_type, return_inverse=True
            )
            return atom_type_index
        except AttributeError:
            return None

    @property
    def atom_name(self) -> np.ndarray:
        if hasattr(self.atoms, "names"):
            return self.atoms.names
        else:
            return np.zeros(self.n_atoms)

    @property
    def atom_name_num(self) -> np.ndarray:
        if hasattr(self.atoms, "names"):
            return np.array(
                list(map(lambda x: data.atom_names.get(x, -1), self.atom_name))
            )
        else:
            return np.repeat(-1, self.n_atoms)

    @property
    def is_nucleic(self) -> np.ndarray:
        return self.bool_selection(self.atoms, "nucleic")

    @property
    def is_peptide(self) -> np.ndarray:
        return self.bool_selection(self.atoms, "protein or (name BB SC*)")

    @property
    def is_lipid(self) -> np.ndarray:
        return np.isin(self.atoms.resnames, data.lipid_names)

    @property
    def is_backbone(self) -> np.ndarray:
        return self.bool_selection(self.atoms, "backbone or nucleicbackbone or name BB")

    @property
    def is_alpha_carbon(self) -> np.ndarray:
        return self.bool_selection(self.atoms, "name CA or name BB")

    @property
    def is_solvent(self) -> np.ndarray:
        return self.bool_selection(
            self.atoms, "name OW or name HW1 or name HW2 or resname W or resname PW"
        )

    @property
    def _attributes_2_blender(self):
        """
        The attributes that will be added to the Blender object.
        """
        return {
            "atomic_number": {
                "value": self.atomic_number,
                "type": "INT",
                "domain": "POINT",
            },
            "vdw_radii": {
                "value": self.vdw_radii,
                "type": "FLOAT",
                "domain": "POINT",
            },
            "mass": {
                "value": self.mass,
                "type": "FLOAT",
                "domain": "POINT",
            },
            "res_id": {
                "value": self.res_id,
                "type": "INT",
                "domain": "POINT",
            },
            "segid": {
                "value": self.segindices,
                "type": "INT",
                "domain": "POINT",
            },
            "res_name": {
                "value": self.res_num,
                "type": "INT",
                "domain": "POINT",
            },
            "b_factor": {
                "value": self.b_factor,
                "type": "FLOAT",
                "domain": "POINT",
            },
            "chain_id": {
                "value": self.chain_id_num,
                "type": "INT",
                "domain": "POINT",
            },
            "atom_types": {
                "value": self.atom_type_num,
                "type": "INT",
                "domain": "POINT",
            },
            "atom_name": {
                "value": self.atom_name_num,
                "type": "INT",
                "domain": "POINT",
            },
            "is_backbone": {
                "value": self.is_backbone,
                "type": "BOOLEAN",
                "domain": "POINT",
            },
            "is_alpha_carbon": {
                "value": self.is_alpha_carbon,
                "type": "BOOLEAN",
                "domain": "POINT",
            },
            "is_solvent": {
                "value": self.is_solvent,
                "type": "BOOLEAN",
                "domain": "POINT",
            },
            "is_nucleic": {
                "value": self.is_nucleic,
                "type": "BOOLEAN",
                "domain": "POINT",
            },
            "is_lipid": {
                "value": self.is_lipid,
                "type": "BOOLEAN",
                "domain": "POINT",
            },
            "is_peptide": {
                "value": self.is_peptide,
                "type": "BOOLEAN",
                "domain": "POINT",
            },
        }

    def save_filepaths_on_object(self) -> None:
        obj = self.object
        obj.mn.filepath_topology = str(path_resolve(self.universe.filename))
        obj.mn.filepath_trajectory = str(
            path_resolve(self.universe.trajectory.filename)
        )

    def reset_playback(self) -> None:
        "Set the playback settings to their default values"
        self.subframes = 0
        self.offset = 0
        self.average = 0
        self.correct_periodic = False
        self.interpolate = False

    def _create_object(
        self, style: str = "vdw", name: str = "NewUniverseObject"
    ) -> bpy.types.Object:
        self.object = databpy.create_object(
            name=name,
            collection=coll.mn(),
            vertices=self.univ_positions,
            edges=self.bonds,
        )

        for att_name, att in self._attributes_2_blender.items():
            try:
                self.store_named_attribute(
                    data=att["value"],
                    name=att_name,
                    atype=att["type"],
                    domain=att["domain"],
                )
            except Exception as e:
                print(e)

        if hasattr(self.atoms, "segindices"):
            segs = []
            for seg in self.atoms.segments:
                segs.append(seg.atoms[0].segid)

            self.object["segments"] = segs

        if style is not None:
            nodes.create_starting_node_tree(
                self.object, style=style, name=f"MN_{self.name}"
            )

    def create_object(
        self,
        name: str = "NewUniverseObject",
        style: str = "vdw",
    ):
        self._create_object(style=style, name=name)

        self.object["chain_ids"] = self.chain_ids

        if hasattr(self, "atom_type_unique"):
            self.object["atom_type_unique"] = self.atom_type_unique

        self.object.mn.entity_type = self._entity_type.value
        self.save_filepaths_on_object()
        bpy.context.view_layer.objects.active = self.object

        return self.object

    def _update_calculations(self):
        for name, func in self.calculations.items():
            try:
                self.store_named_attribute(data=func(self.universe), name=name)
            except Exception as e:
                print(e)

    def _update_selections(self):
        objs_to_update = [obj for obj in bpy.data.objects if obj.uuid == self.uuid]

        # mark all selections for cleanup if they are no longer relevant
        for selection in self.selections.values():
            selection.cleanup = True

        for obj in objs_to_update:
            for sel in obj.mn_trajectory_selections:
                # try and get a corresponding selection for this named selection
                # if the selection can't be found we create one
                selection = self.selections.get(sel.name)
                if selection is None:
                    selection = self.selection_from_ui(sel)
                elif sel.updating:
                    # if the selection string has, or some of the parameters about it have
                    # changed then we have to change the selection's AtomGroup before we
                    # apply the selection to the mesh
                    if (
                        selection.selection_str != sel.selection_str
                        or selection.periodic != sel.periodic
                    ):
                        selection.change_selection(
                            selection_str=sel.selection_str,
                            name=sel.name,
                            updating=sel.updating,
                            periodic=sel.periodic,
                        )

                    # Apply the selection to the actual mesh in the form of a boolean
                    # named attribute
                    self.apply_selection(selection)

                # mark the selection to not be cleaned up, and add any message from the
                # selection to the UI selection item for display in the UI
                selection.cleanup = False
                sel.message = selection.message

        # remove all of the attributes and selections that are marked for cleanup because
        # they are no longer being used by any objects for selections
        for name in list(self.selections):
            if self.selections[name].cleanup:
                try:
                    self.object.data.attributes.remove(
                        self.object.data.attributes[name]
                    )
                    self.selections.pop(name)
                except Exception as e:
                    print(e)

    @property
    def _frame(self) -> int:
        return self.object.mn.frame_hidden

    @_frame.setter
    def _frame(self, value: int) -> None:
        self.object.mn.frame_hidden = value

    @property
    def frame(self) -> int:
        return self.object.mn.frame

    @frame.setter
    def frame(self, value: int) -> None:
        self.object.mn.frame = value

    @property
    def update_with_scene(self) -> bool:
        return self.object.mn.update_with_scene

    @update_with_scene.setter
    def update_with_scene(self, value: bool) -> None:
        self.object.mn.update_with_scene = value

    @property
    def subframes(self) -> int:
        return self.object.mn.subframes

    @subframes.setter
    def subframes(self, value: int) -> None:
        self.object.mn.subframes = value

    @property
    def offset(self) -> int:
        return self.object.mn.offset

    @offset.setter
    def offset(self, value: int) -> None:
        self.object.mn.offset = value

    @property
    def average(self) -> int:
        return self.object.mn.average

    @average.setter
    def average(self, value: int) -> None:
        self.object.mn.average = value

    @property
    def correct_periodic(self) -> bool:
        return self.object.mn.correct_periodic

    @correct_periodic.setter
    def correct_periodic(self, value: bool) -> None:
        self.object.mn.correct_periodic = value

    @property
    def interpolate(self) -> bool:
        return self.object.mn.interpolate

    @interpolate.setter
    def interpolate(self, value: bool) -> None:
        self.object.mn.interpolate = value

    def _frame_range(self, frame: int):
        "Get the trajectory frame numbers over which we will average values"
        return frames_to_average(frame, self.average)

    def _cache_ordered(self) -> np.ndarray:
        "Return the cached frames as a 3D array, in chronological order"
        keys = list(self.cache.keys())
        keys.sort()
        return np.array([self.cache[k] for k in keys])

    def adjust_periodic_positions(
        self, pos1: np.ndarray, pos2: np.ndarray
    ) -> np.ndarray:
        "Returns the input pos2 with a periodic correction potentially applied"
        if self.correct_periodic and self.is_orthorhombic:
            return correct_periodic_positions(pos1, pos2, self.universe.dimensions[:3])
        else:
            return pos2

    def position_cache_mean(self, frame: int) -> np.ndarray:
        "Return the mean position from the currently cached positions"
        self.update_position_cache(frame)

        if self.average == 0:
            return self.cache[frame]

        array = self._cache_ordered()
        if self.correct_periodic and self.is_orthorhombic:
            # we want to correct the periodic boundary crossing in refernce to the fist
            # frame we are averaging
            for i, pos in enumerate(array):
                if i == 0:
                    continue
                array[i] = self.adjust_periodic_positions(array[0], pos)

        return np.mean(array, axis=0)

    def set_frame(self, frame: int) -> None:
        """
        Update the positions, selections and calculations for this trajectory, based on
        frame number of the current scene, not the frame number of the Universe
        """
        self._frame = self.frame_mapper(frame)
        self._update_positions(frame)
        self._update_selections()
        self._update_calculations()

    def _position_at_frame(self, frame: int) -> np.ndarray:
        "Return the atom positions at the given universe frame number"
        self.uframe = frame
        return self.univ_positions

    def update_position_cache(self, frame: int, cache_ahead: bool = True) -> None:
        "Update the currently cached positions, based on the new frame"
        # get the individual frame numbers that we will be caching
        frames_to_cache = self._frame_range(frame)

        # if we should be looking ahead by 1 for interpolating, ensure we are caching 1
        # frame ahead so when the frame changes we already have it stored and aren't
        # double dipping
        if len(frames_to_cache) == 1 and cache_ahead:
            frames_to_cache = np.array(
                (frames_to_cache[0], frames_to_cache[0] + 1), dtype=int
            )

        # only cleanup the cache if we have more than 2 frame stored, helps when moving
        # forward or back a single frame
        if len(self.cache) > 2:
            # remove any frames that no longer need to be cached
            to_remove = [f for f in self.cache if f not in frames_to_cache]
            for f in to_remove:
                del self.cache[f]

        # update the cache with any frames that are not yet cached
        for f in frames_to_cache:
            if f not in self.cache:
                self.cache[f] = self._position_at_frame(f)

    def frame_mapper(self, frame: int):
        return frame_mapper(
            frame=frame,
            subframes=self.subframes,
            offset=self.offset,
            mapping=self.frame_mapping,
        )

    def _update_positions(self, frame):
        """
        The function that will be called when the frame changes.
        It will update the positions and selections of the atoms in the scene.
        """
        # get the two frames of the trajectory to potentially access data from
        # uframe_current, uframe_next = [self.frame_mapper(x) for x in (frame, frame + 1)]
        uframe_current = self.frame_mapper(frame)
        uframe_next = uframe_current + 1

        if self.subframes > 0 and self.interpolate:
            # if we are adding subframes and interpolating, then we get the positions
            # at the two universe frames, then interpolate between them, potentially
            # correcting for any periodic boundary crossing
            pos_current = self.position_cache_mean(
                uframe_current,
            )
            pos_next = self.position_cache_mean(uframe_next)

            # if we are averaging, then we have already applied periodic correction
            # and we can skip this step
            if self.correct_periodic and self.is_orthorhombic and self.average == 0:
                pos_next = correct_periodic_positions(
                    pos_current,
                    pos_next,
                    dimensions=self.universe.dimensions[:3] * self.world_scale,
                )

            # interpolate between the two sets of positions
            self.position = databpy.lerp(
                pos_current, pos_next, t=fraction(frame, self.subframes + 1)
            )
        elif self.average > 0:
            # if we have subframes then we get the potential mean positions for the cached
            # frames that we are looking at
            self.position = self.position_cache_mean(uframe_current)
        else:
            # otherwise just get the current positions for the relevant frame and set
            # those on the object
            self.position = self._position_at_frame(uframe_current)

    def __repr__(self):
        return f"<Trajectory, `universe`: {self.universe}, `object`: {self.object}"



import os
import shutil

import bpy

from .utils import ADDON_DIR

SUBFOLDER = "Molecular Nodes"


def is_installed():
    base_path = bpy.utils.user_resource(
        "SCRIPTS", path=os.path.join("startup", "bl_app_templates_user"), create=False
    )
    molecular_nodes_path = os.path.join(base_path, "Molecular Nodes")
    return os.path.exists(molecular_nodes_path)


def install():
    # Construct the base path for app templates
    base_path = bpy.utils.user_resource(
        "SCRIPTS", path=os.path.join("startup", "bl_app_templates_user"), create=True
    )

    # Ensure the base path exists
    if not os.path.exists(base_path):
        os.makedirs(base_path)

    # Construct the full path to the "Molecular Nodes" subfolder
    path_app_templates = os.path.join(base_path, SUBFOLDER)

    # Create the "Molecular Nodes" folder if it doesn't exist
    if not os.path.exists(path_app_templates):
        os.makedirs(path_app_templates)

    # Define the path to the startup file
    startup_file = os.path.join(
        os.path.abspath(ADDON_DIR), "assets", "template", "startup.blend"
    )

    # Copy the startup file to the "Molecular Nodes" folder
    shutil.copy(startup_file, path_app_templates)

    # Refresh Blender's script paths to recognize the new template
    bpy.utils.refresh_script_paths()


def uninstall():
    for folder in bpy.utils.app_template_paths():
        path = os.path.join(os.path.abspath(folder), SUBFOLDER)
        if "Molecular Nodes" not in path:
            continue

        if os.path.exists(path):
            shutil.rmtree(path)
    bpy.utils.refresh_script_paths()



from abc import ABCMeta
import os
import numpy as np
import bpy
from typing import Union
from pathlib import Path
from ..entity import MolecularEntity, EntityType
from ... import blender as bl
import databpy


class Density(MolecularEntity, metaclass=ABCMeta):
    """
    Abstract base class for molecular density objects.

    """

    def __init__(self, file_path: Union[str, Path]):
        super().__init__()
        self._entity_type = EntityType.DENSITY
        self.file_path: Path = bl.path_resolve(file_path)
        self.grid = None
        self.file_vdb: str
        self.threshold: float

    def named_attribute(self, name: str, evaluate: bool = True) -> np.ndarray:
        obj = bl.mesh.evaluate_using_mesh(self.object)
        return databpy.named_attribute(obj, name, evaluate=True)

    def path_to_vdb(self, file: str, center: bool = False, invert: bool = False):
        """
        Convert a file path to a corresponding VDB file path.

        Parameters
        ----------
        file : str
            The path of the original file.

        Returns
        -------
        str
            The path of the corresponding VDB file.
        """
        # Set up file paths
        folder_path = os.path.dirname(file)
        name = os.path.basename(file).split(".")[0]
        name += "_center" if center else ""
        name += "_invert" if invert else ""
        file_name = name + ".vdb"
        file_path = os.path.join(folder_path, file_name)
        return file_path



import os
import sys
import numpy as np

from pathlib import Path
from math import floor
from mathutils import Matrix

ADDON_DIR = Path(__file__).resolve().parent
MN_DATA_FILE = os.path.join(ADDON_DIR, "assets", "MN_data_file_4.2.blend")


def add_current_module_to_path():
    path = str(ADDON_DIR.parent)
    sys.path.append(path)


def fraction(x, y):
    return x % y / y


def correct_periodic_1d(
    value1: np.ndarray, value2: np.ndarray, boundary: float
) -> np.ndarray:
    diff = value2 - value1
    half = boundary / 2
    value2[diff > half] -= boundary
    value2[diff < -half] += boundary
    return value2


def correct_periodic_positions(
    positions_1: np.ndarray, positions_2: np.ndarray, dimensions: np.ndarray
) -> np.ndarray:
    if not np.allclose(dimensions[3:], 90.0):
        raise ValueError(
            f"Only works with orthorhombic unitcells, and not dimensions={dimensions}"
        )
    final_positions = positions_2.copy()
    for i in range(3):
        final_positions[:, i] = correct_periodic_1d(
            positions_1[:, i], positions_2[:, i], dimensions[i]
        )
    return final_positions


def frame_mapper(
    frame: int,
    subframes: int = 0,
    offset: int = 0,
    mapping: np.ndarray | None = None,
) -> int:
    frame = max(frame - offset, 0)

    if mapping is not None:
        if not isinstance(mapping, np.ndarray):
            raise ValueError(
                "Frame mapping must be an array of values to map frames to"
            )
        # add the subframes to the frame mapping
        frame_map = np.repeat(mapping, subframes + 1)
        # get the current and next frames
        frame_a = frame_map[frame]

    frame_a = frame

    if subframes > 0:
        frame_a = int(frame / (subframes + 1))

    return frame_a


def frames_to_average(frame: int, average: int = 0, lower_bound: int = 0) -> np.ndarray:
    length = average * 2 + 1
    frames = np.arange(length) + frame - average
    frames = frames[frames >= lower_bound]
    return frames


# data types for the np.array that will store per-chain symmetry operations
dtype = [
    ("assembly_id", int),
    ("transform_id", int),
    ("chain_id", "U10"),
    ("rotation", float, 4),  # quaternion form
    ("translation", float, 3),
]


def array_quaternions_from_dict(transforms_dict):
    n_transforms = 0
    for assembly in transforms_dict.values():
        for transform in assembly:
            n_transforms += len(transform[0])

    arr = np.array((n_transforms), dtype=dtype)

    transforms = []
    for i, assembly in enumerate(transforms_dict.values()):
        for j, transform in enumerate(assembly):
            chains = transform[0]
            matrix = transform[1]
            arr = np.zeros((len(chains)), dtype=dtype)
            translation, rotation, scale = Matrix(matrix).decompose()
            arr["assembly_id"] = i + 1
            arr["transform_id"] = j
            arr["chain_id"] = chains
            arr["rotation"] = rotation
            arr["translation"] = translation
            transforms.append(arr)

    return np.hstack(transforms)



import bpy
from .node_info import menu_items


class MN_MT_Node_Color(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_COLOR"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("color").menu(self.layout, context)


class MN_MT_Node_Bonds(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_BONDS"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("bonds").menu(self.layout, context)


class MN_MT_Node_Style(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_STYLE"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("style").menu(self.layout, context)


class MN_MT_Node_Select(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_SELECT"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("select").menu(self.layout, context)


class MN_MT_Node_Assembly(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_ASSEMBLY"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("ensemble").menu(self.layout, context)


class MN_MT_Node_DNA(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_DNA"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("DNA").menu(self.layout, context)


class MN_MT_Node_Animate(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_ANIMATE"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("animate").menu(self.layout, context)


class MN_MT_Node_Density(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_DENSITY"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("density").menu(self.layout, context)


class MN_MT_Node_Topology(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_TOPOLOGY"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("topology").menu(self.layout, context)


class MN_MT_Node_Curves(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_CURVES"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("curves").menu(self.layout, context)


class MN_MT_Node_Geometry(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_GEOMETRY"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("geometry").menu(self.layout, context)


class MN_MT_Node_Utils(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_UTILS"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("utils").menu(self.layout, context)


class MN_MT_Node_Fields(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_FIELDS"
    bl_label = ""

    def draw(self, context):
        menu_items.get_submenu("fields").menu(self.layout, context)


def draw_node_menus(self, context):
    layout = self.layout
    layout.menu("MN_MT_NODE_STYLE", text="Style", icon_value=77)
    layout.menu("MN_MT_NODE_SELECT", text="Select", icon="RESTRICT_SELECT_OFF")
    layout.menu("MN_MT_NODE_COLOR", text="Color", icon="COLORSET_07_VEC")
    layout.separator()
    layout.menu("MN_MT_NODE_ANIMATE", text="Animation", icon="MOD_DASH")
    layout.menu("MN_MT_NODE_GEOMETRY", text="Geometry", icon="MESH_DATA")
    layout.menu("MN_MT_NODE_TOPOLOGY", text="Topology", icon="ORIENTATION_CURSOR")
    layout.menu("MN_MT_NODE_ASSEMBLY", text="Ensemble", icon="GROUP_VERTEX")
    layout.separator()
    layout.menu("MN_MT_NODE_DENSITY", text="Density", icon="VOLUME_DATA")
    layout.separator()
    layout.menu("MN_MT_NODE_DNA", text="DNA", icon="GP_SELECT_BETWEEN_STROKES")
    layout.separator()
    layout.menu("MN_MT_NODE_CURVES", text="Curves", icon="CURVE_DATA")
    layout.menu("MN_MT_NODE_UTILS", text="Utilities", icon="TOOL_SETTINGS")
    layout.menu("MN_MT_NODE_FIELDS", text="Fields", icon="NODE")


class MN_MT_Node(bpy.types.Menu):
    bl_idname = "MN_MT_NODE"
    bl_label = "Menu for Adding Nodes in GN Tree"

    def draw(self, context):
        draw_node_menus(self, context)


def MN_add_node_menu(self, context):
    if "GeometryNodeTree" == bpy.context.area.spaces[0].tree_type:
        layout = self.layout
        layout.menu("MN_MT_NODE", text="Molecular Nodes", icon="PARTICLE_DATA")


CLASSES = [
    MN_MT_Node,
    MN_MT_Node_Animate,
    MN_MT_Node_Assembly,
    MN_MT_Node_Bonds,
    MN_MT_Node_Color,
    MN_MT_Node_Density,
    MN_MT_Node_DNA,
    MN_MT_Node_Style,
    MN_MT_Node_Select,
    MN_MT_Node_Topology,
    MN_MT_Node_Geometry,
    MN_MT_Node_Curves,
    MN_MT_Node_Utils,
    MN_MT_Node_Fields,
]



import io
import os
from pathlib import Path
import requests

CACHE_OLD = str(Path("~", ".MolecularNodes").expanduser())
CACHE_DIR = str(Path("~", "MolecularNodesCache").expanduser())

# rename old cache directories if users have them so we aren't leaving cached files in
# hidden folders on disk somewhere, I don't like the idea of silently renaming folders
# on a user's disk on load, so for now this will be disabled.
# TODO: make a decision on this (maybe a conformation popup on download)
# if os.path.exists(CACHE_OLD):
#     os.rename(CACHE_OLD, CACHE_DIR)


class FileDownloadPDBError(Exception):
    """
    Exception raised for errors in the file download process.

    Attributes:
        message -- explanation of the error
    """

    def __init__(
        self,
        message="There was an error downloading the file from the Protein Data Bank. PDB or format for PDB code may not be available.",
    ):
        self.message = message
        super().__init__(self.message)


def download(code, format="cif", cache=CACHE_DIR, database="rcsb"):
    """
    Downloads a structure from the specified protein data bank in the given format.

    Parameters
    ----------
    code : str
        The code of the file to fetch.
    format : str, optional
        The format of the file. Defaults to "cif". Possible values are ['cif', 'pdb',
        'mmcif', 'pdbx', 'bcif'].
    cache : str, optional
        The cache directory to store the fetched file. Defaults to `~/MolecularNodesCache`.
    database : str, optional
        The database to fetch the file from. Defaults to 'rcsb'.

    Returns
    -------
    file
        The fetched file as a file-like object.

    Raises
    ------
    ValueError
        If the specified format is not supported.
    """
    supported_formats = ["cif", "pdb", "bcif"]
    if format not in supported_formats:
        raise ValueError(f"File format '{format}' not in: {supported_formats=}")

    _is_binary = format in ["bcif"]
    filename = f"{code}.{format}"
    # create the cache location
    if cache:
        if not os.path.isdir(cache):
            os.makedirs(cache)

        file = os.path.join(cache, filename)
    else:
        file = None

    if file:
        if os.path.exists(file):
            return file

    # get the contents of the url
    try:
        r = requests.get(_url(code, format, database))
        r.raise_for_status()
    except requests.HTTPError:
        raise FileDownloadPDBError
    if _is_binary:
        content = r.content
    else:
        content = r.text

    if file:
        mode = "wb+" if _is_binary else "w+"
        with open(file, mode) as f:
            f.write(content)
    else:
        if _is_binary:
            file = io.BytesIO(content)
        else:
            file = io.StringIO(content)

    return file


def _url(code, format, database="rcsb"):
    "Get the URL for downloading the given file form a particular database."

    if database in ["rcsb", "pdb", "wwpdb"]:
        if format == "bcif":
            return f"https://models.rcsb.org/{code}.bcif"

        else:
            return f"https://files.rcsb.org/download/{code}.{format}"
    # if database == "pdbe":
    #     return f"https://www.ebi.ac.uk/pdbe/entry-files/download/{filename}"
    elif database == "alphafold":
        return get_alphafold_url(code, format)
    # if database == "pdbe":
    #     return f"https://www.ebi.ac.uk/pdbe/entry-files/download/{filename}"
    else:
        ValueError(f"Database {database} not currently supported.")


def get_alphafold_url(code, format):
    if format not in ["pdb", "cif", "bcif"]:
        ValueError(f"Format {format} not currently supported from AlphaFold databse.")

    # we have to first query the database, then they'll return some JSON with a list
    # of metadata, some items of which will be the URLs for the computed models
    # in different formats such as pdbUrl, cifUrl, bcifUrl
    url = f"https://alphafold.ebi.ac.uk/api/prediction/{code}"
    response = requests.get(url)
    data = response.json()[0]
    return data[f"{format}Url"]



import bpy
from pathlib import Path


def path_resolve(path: str | Path) -> Path:
    if isinstance(path, str):
        return Path(bpy.path.abspath(path))
    elif isinstance(path, Path):
        return Path(bpy.path.abspath(str(path)))
    else:
        raise ValueError(f"Unable to resolve path: {path}")



from abc import ABCMeta
import bpy
from enum import Enum
from databpy import (
    BlenderObject,
)


# create a EntityType enum with strings for values, "md", "md-oxdna", "molecule", "star"
class EntityType(Enum):
    MD = "md"
    MD_OXDNA = "md-oxdna"
    MOLECULE = "molecule"
    STAR = "star"
    DENSITY = "density"


class MolecularEntity(
    BlenderObject,
    metaclass=ABCMeta,
):
    def __init__(self) -> None:
        super().__init__(obj=None)
        self._entity_type: EntityType
        self._register_with_session()

    @property
    def bob(self) -> BlenderObject:
        return BlenderObject(self.object)

    @property
    def node_group(self) -> bpy.types.NodeGroup:
        return self.object.modifiers["MolecularNodes"].node_group

    def _register_with_session(self) -> None:
        bpy.context.scene.MNSession.register_entity(self)

    def set_frame(self, frame: int) -> None:
        """
        Update the underlying data to correspond to changes in the scene frame.

        This method should be implemented by subclasses to update the entity's
        data based on the given frame number. This can include updating positions
        and performing other necessary calculations.

        Args:
            frame (int): The frame number to update the entity's data to.

        Raises:
            NotImplementedError: If the method is not implemented by a subclass.
        """
        raise NotImplementedError("Subclasses must implement this method")



import bpy
from abc import ABCMeta
from ... import blender as bl
from ..entity import MolecularEntity
from typing import Union
from pathlib import Path


class Ensemble(MolecularEntity, metaclass=ABCMeta):
    def __init__(self, file_path: Union[str, Path]):
        """
        Initialize an Ensemble object.

        Parameters
        ----------
        file_path : str
            The path to the file.

        """
        super().__init__()
        self.type: str = "ensemble"
        self.file_path: Path = bl.path_resolve(file_path)

    @property
    def instance_collection(self) -> bpy.types.Collection:
        """
        The instances of the ensemble.

        """
        return bpy.data.collections[self._instance_collection_name]

    @instance_collection.setter
    def instance_collection(self, value: bpy.types.Collection) -> None:
        if not isinstance(value, bpy.types.Collection):
            raise ValueError("The instances must be a bpy.types.Collection.")
        self._instance_collection_name = value.name

    def create_object(
        self,
        name: str = "NewEnsemble",
        node_setup: bool = True,
        world_scale: float = 0.01,
        fraction: float = 1.0,
        simplify=False,
    ):
        """
        Create a 3D object for the ensemble.

        Parameters
        ----------
        name : str, optional
            The name of the model (default is "NewEnsemble").
        node_setup : bool, optional
            Whether to setup nodes for the data and instancing objects. (default is True).
        world_scale : float, optional
            Scaling transform for the coordinates before loading in to Blender. (default is 0.01).
        fraction : float, optional
            The fraction of the instances to display on loading. Reducing can help with performance. (default is 1.0).
        simplify : bool, optional
            Whether to isntance the given models or simplify them for debugging and performance. (default is False).

        Creates a data object which stores all of the required instancing information. If
        there are molecules to be instanced, they are also created in their own data collection.

        Parameters:
        - name (str): The name of the model. Default is "NewEnsemble".
        - node_setup (bool): Whether to set up nodes. Default is True.
        - world_scale (float): The scale of the world. Default is 0.01.
        - fraction (float): The fraction of molecules to be instanced. Default is 1.0.
        - simplify (bool): Whether to simplify the model. Default is False.

        """
        pass



import bpy

from .. import __package__, template


class MN_OT_Template_Install(bpy.types.Operator):
    bl_idname = "mn.template_install"
    bl_label = "Install Template"
    bl_description = "Install the Molecular Nodes startup template file."

    def execute(self, context):
        template.install()
        self.report({"INFO"}, "Installed Molecular Nodes template.")
        return {"FINISHED"}


class MN_OT_Template_Uninstall(bpy.types.Operator):
    bl_idname = "mn.template_uninstall"
    bl_label = "Uninstall Template"
    bl_description = "Uninstall the Molecular Nodes startup template file."

    @classmethod
    def poll(cls, context):
        return template.is_installed()

    def execute(self, context):
        try:
            template.uninstall()
            self.report({"INFO"}, "Uninstalled Molecular Nodes template.")
        except FileNotFoundError:
            self.report({"WARNING"}, "Template not installed.")

        return {"FINISHED"}


class MolecularNodesPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    def draw(self, context):
        layout = self.layout
        layout.label(
            text="Install the Molecular Nodes template file, to start Blender with useful default settings"
        )
        row = layout.row()
        if not template.is_installed():
            text = "Install Template"
        else:
            text = "Reinstall Template"

        row.operator("mn.template_install", text=text)
        row.operator("mn.template_uninstall")


CLASSES = [MN_OT_Template_Install, MN_OT_Template_Uninstall, MolecularNodesPreferences]



import bpy
from .mrc import MRC

bpy.types.Scene.MN_import_density_invert = bpy.props.BoolProperty(
    name="Invert Data",
    description="Invert the values in the map. Low becomes high, high becomes low.",
    default=False,
)
bpy.types.Scene.MN_import_density_center = bpy.props.BoolProperty(
    name="Center Density",
    description="Translate the density so that the center of the box is at the origin.",
    default=False,
)
bpy.types.Scene.MN_import_density = bpy.props.StringProperty(
    name="File",
    description="File path for the map file.",
    subtype="FILE_PATH",
    maxlen=0,
)
bpy.types.Scene.MN_import_density_name = bpy.props.StringProperty(
    name="Name",
    description="Name for the new density object.",
    default="NewDensityObject",
    maxlen=0,
)

bpy.types.Scene.MN_import_density_style = bpy.props.EnumProperty(
    name="Style",
    items=(
        (
            "density_surface",
            "Surface",
            "A mesh surface based on the specified threshold",
            0,
        ),
        (
            "density_wire",
            "Wire",
            "A wire mesh surface based on the specified threshold",
            1,
        ),
    ),
)


def load(
    file_path: str,
    name: str = "NewDensity",
    invert: bool = False,
    setup_nodes: bool = True,
    style: str = "density_surface",
    center: bool = False,
    overwrite: bool = False,
):
    density = MRC(
        file_path=file_path, center=center, invert=invert, overwrite=overwrite
    )
    density.create_object(name=name, setup_nodes=setup_nodes, style=style)
    return density


class MN_OT_Import_Map(bpy.types.Operator):
    bl_idname = "mn.import_density"
    bl_label = "Load"
    bl_description = "Import a EM density map into Blender"
    bl_options = {"REGISTER"}

    def execute(self, context):
        scene = context.scene
        load(
            file_path=scene.MN_import_density,
            name=scene.MN_import_density_name,
            invert=scene.MN_import_density_invert,
            setup_nodes=scene.MN_import_node_setup,
            style=scene.MN_import_density_style,
            center=scene.MN_import_density_center,
        )
        return {"FINISHED"}


def panel(layout, scene):
    layout.label(text="Load EM Map", icon="FILE_TICK")
    layout.separator()

    row = layout.row()
    row.prop(scene, "MN_import_density_name")
    row.operator("mn.import_density")

    layout.prop(scene, "MN_import_density")
    layout.separator()
    col = layout.column()
    col.alignment = "LEFT"
    col.scale_y = 0.5
    label = f"\
    An intermediate file will be created: {scene.MN_import_density}.vdb\
    Please do not delete this file or the volume will not render.\
    Move the original .map file to change this location.\
    "
    for line in label.strip().split("    "):
        col.label(text=line)

    layout.separator()
    layout.label(text="Options", icon="MODIFIER")

    row = layout.row()
    row.prop(scene, "MN_import_node_setup", text="")
    col = row.column()
    col.prop(scene, "MN_import_density_style")
    col.enabled = scene.MN_import_node_setup

    grid = layout.grid_flow()
    grid.prop(scene, "MN_import_density_invert")
    grid.prop(scene, "MN_import_density_center")



import io
import os
import tempfile

import biotite.database.rcsb as rcsb
import pytest
from biotite.structure.io import load_structure

import molecularnodes as mn
from molecularnodes.download import FileDownloadPDBError, download

from .constants import codes

# currently can't figure out downloading from other services
DATABASES = ["rcsb"]


def _filestart(format):
    if format == "cif":
        return "data_"
    else:
        return "HEADER"


def test_download_raises_error_on_invalid_format():
    with pytest.raises(ValueError) as excinfo:
        download("1abc", "invalid_format")
    assert (
        "File format 'invalid_format' not in: supported_formats=['cif', 'pdb', 'bcif']"
        in str(excinfo.value)
    )


def test_fail_download_pdb_large_structure_raises():
    with pytest.raises(FileDownloadPDBError) as excinfo:
        download("7D6Z", format="pdb")

    assert (
        "There was an error downloading the file from the Protein Data Bank. PDB or format for PDB code may not be available."
        in str(excinfo.value)
    )


@pytest.mark.parametrize("format", ["cif", "bcif", "pdb"])
def test_compare_biotite(format):
    struc_download = load_structure(
        download("4ozs", format=format, cache=tempfile.TemporaryDirectory().name)
    )
    struc_biotite = load_structure(
        rcsb.fetch(
            "4ozs", format=format, target_path=tempfile.TemporaryDirectory().name
        )
    )
    assert struc_download == struc_biotite


@pytest.mark.parametrize("code", codes)
@pytest.mark.parametrize("database", DATABASES)
@pytest.mark.parametrize("format", ["pdb", "cif"])
def test_fetch_with_cache(tmpdir, code, format, database):
    cache_dir = tmpdir.mkdir("cache")
    file = download(code, format, cache=str(cache_dir), database=database)

    assert isinstance(file, str)
    assert os.path.isfile(file)
    assert file.endswith(f"{code}.{format}")

    with open(file, "r") as f:
        content = f.read()
    assert content.startswith(_filestart(format))


DATABASES = ["rcsb"]  # currently can't figure out downloading from the pdbe


@pytest.mark.parametrize("code", codes)
@pytest.mark.parametrize("database", DATABASES)
@pytest.mark.parametrize("format", ["pdb", "cif"])
def test_fetch_without_cache(tmpdir, code, format, database):
    file = download(code, format, cache=None, database=database)

    assert isinstance(file, io.StringIO)
    content = file.getvalue()
    assert content.startswith(_filestart(format))


@pytest.mark.parametrize("database", DATABASES)
def test_fetch_with_invalid_format(database):
    code = "4OZS"
    format = "xyz"

    with pytest.raises(ValueError):
        download(code, format, cache=None, database=database)


@pytest.mark.parametrize("code", codes)
@pytest.mark.parametrize("database", DATABASES)
@pytest.mark.parametrize("format", ["bcif"])
def test_fetch_with_binary_format(tmpdir, code, database, format):
    cache_dir = tmpdir.mkdir("cache")
    file = download(code, format, cache=str(cache_dir), database=database)

    assert isinstance(file, str)
    assert os.path.isfile(file)
    assert file.endswith(f"{code}.{format}")

    if format == "bcif":
        start = b"\x83\xa7"

    with open(file, "rb") as f:
        content = f.read()
    assert content.startswith(start)


# TODO BCIF is supported elsewhere in the package but can't currently be parsed properly
# I think there is something weird going on with the alphafold formatted bcif files


@pytest.mark.parametrize("format", ("cif", "pdb"))
@pytest.mark.parametrize("code", ("A0A5E8G9H8", "A0A5E8G9T8", "K4PA18"))
def test_alphafold_download(format: str, code: str, tmpdir) -> None:
    file = download(code=code, format=format, database="alphafold", cache=tmpdir)

    mol = mn.entities.load_local(file)

    assert mol.array



import bpy
from bpy.props import StringProperty, BoolProperty
from ...handlers import _selection_update_trajectories, _update_trajectories


class TrajectorySelectionItem(bpy.types.PropertyGroup):
    """Group of properties for custom selections for MDAnalysis import."""

    name: StringProperty(  # type: ignore
        name="Name",
        description="Name of the attribute on the mesh",
        default="custom_selection",
        update=_selection_update_trajectories,
    )

    selection_str: StringProperty(  # type: ignore
        name="Selection",
        description="Selection to be applied, written in the MDAnalysis selection language",
        default="name CA",
        update=_selection_update_trajectories,
    )

    updating: BoolProperty(  # type: ignore
        name="Updating",
        description="Recalculate the selection on scene frame change",
        default=True,
        update=_selection_update_trajectories,
    )

    periodic: BoolProperty(  # type: ignore
        name="Periodic",
        description="For geometric selections, whether to account for atoms in different periodic images when searching",
        default=True,
        update=_selection_update_trajectories,
    )

    message: StringProperty(  # type: ignore
        name="Message",
        description="Message to report back from `universe.select_atoms()`",
        default="",
    )

    immutable: BoolProperty(  # type: ignore
        name="Immutable",
        description="Whether the selection is immutable",
        default=False,
    )


class MN_UL_TrajectorySelectionListUI(bpy.types.UIList):
    """UI List"""

    def draw_item(
        self, context, layout, data, item, icon, active_data, active_propname, index
    ):
        custom_icon = "VIS_SEL_11"

        if self.layout_type in {"DEFAULT", "COMPACT"}:
            row = layout.row()
            if item.message != "":
                custom_icon = "ERROR"
                row.alert = True

            row.prop(item, "name", text="", emboss=False)
            row.prop(item, "updating", icon_only=True, icon="FILE_REFRESH")
            row.prop(item, "periodic", icon_only=True, icon="CUBE")
            if item.immutable:
                row.enabled = False

        elif self.layout_type in {"GRID"}:
            layout.alignment = "CENTER"
            layout.label(text="", icon=custom_icon)


class MN_OT_Universe_Selection_Add(bpy.types.Operator):
    "Add a new custom selection to a trajectory"

    bl_idname = "mn.trajectory_selection_add"
    bl_label = "+"
    bl_description = "Add a new boolean attribute for the given MDA selection string"

    def execute(self, context):
        obj = context.active_object
        obj.mn_trajectory_selections.add()
        i = int(len(obj.mn_trajectory_selections) - 1)
        obj.mn_trajectory_selections[i].name = f"selection_{i + 1}"
        obj.mn["list_index"] = i
        _update_trajectories(self, context)

        return {"FINISHED"}


class MN_OT_Universe_Selection_Delete(bpy.types.Operator):
    bl_idname = "mda.delete_item"
    bl_label = "-"
    bl_description = "Delete the given boolean selection from the universe"

    @classmethod
    def poll(cls, context):
        return context.active_object.mn_trajectory_selections

    def execute(self, context):
        obj = context.active_object
        index = obj.mn.trajectory_selection_index

        sel_list = obj.mn_trajectory_selections
        sel_list.remove(index)
        obj.mn.trajectory_selection_index = len(sel_list) - 1
        _update_trajectories(self, context)

        return {"FINISHED"}


CLASSSES = [
    TrajectorySelectionItem,  # has to be registered before the others to work properly
    MN_UL_TrajectorySelectionListUI,
    MN_OT_Universe_Selection_Add,
    MN_OT_Universe_Selection_Delete,
]



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import bpy
from bpy.app.handlers import frame_change_post, load_post, save_post
from bpy.props import PointerProperty, CollectionProperty
from .handlers import update_trajectories
from . import entities, operators, props, session, ui
from .utils import add_current_module_to_path
from .ui import pref
from .ui.node_menu import MN_add_node_menu
from .ui.panel import MN_PT_Scene, pt_object_context, change_style_node_menu

all_classes = (
    ui.CLASSES
    + operators.CLASSES
    + entities.CLASSES
    + props.CLASSES
    + [
        MN_PT_Scene,
    ]
    + pref.CLASSES
    + session.CLASSES
)


def _test_register():
    try:
        register()
    except Exception as e:
        print(e)
        unregister()
        register()


def register():
    # register all of the import operators
    for op in all_classes:
        try:
            bpy.utils.register_class(op)
        except Exception as e:
            # print(e)
            pass
    add_current_module_to_path()
    bpy.types.NODE_MT_add.append(MN_add_node_menu)
    bpy.types.VIEW3D_MT_object_context_menu.prepend(pt_object_context)
    bpy.types.NODE_MT_context_menu.prepend(change_style_node_menu)

    save_post.append(session._pickle)
    load_post.append(session._load)
    frame_change_post.append(update_trajectories)

    bpy.types.Scene.MNSession = session.MNSession()
    bpy.types.Object.mn = PointerProperty(type=props.MolecularNodesObjectProperties)
    bpy.types.Scene.mn = PointerProperty(type=props.MolecularNodesSceneProperties)
    bpy.types.Object.mn_trajectory_selections = CollectionProperty(
        type=entities.trajectory.props.TrajectorySelectionItem
    )


def unregister():
    for op in all_classes:
        try:
            bpy.utils.unregister_class(op)
        except Exception as e:
            # print(e)
            pass

    bpy.types.NODE_MT_add.remove(MN_add_node_menu)
    bpy.types.VIEW3D_MT_object_context_menu.remove(pt_object_context)
    bpy.types.NODE_MT_context_menu.remove(change_style_node_menu)

    save_post.remove(session._pickle)
    load_post.remove(session._load)
    frame_change_post.remove(update_trajectories)
    del bpy.types.Scene.MNSession
    del bpy.types.Scene.mn
    del bpy.types.Object.mn
    del bpy.types.Object.mn_trajectory_selections



import MDAnalysis as mda
import numpy.typing as npt
import numpy as np


class Selection:
    def __init__(
        self, universe: mda.Universe, selection_str, name, updating=True, periodic=True
    ):
        self.selection_str: str = selection_str
        self.periodic: bool = periodic
        self.updating: bool = updating
        self.universe: mda.Universe = universe
        self.message: str = ""
        self.name: str = name
        self.cleanup: bool = True
        self.ag = universe.select_atoms(
            selection_str, updating=updating, periodic=periodic
        )
        self.mask_array = self._ag_to_mask()

    def _ag_to_mask(self) -> npt.NDArray[np.bool_]:
        "Return a 1D boolean mask for the Universe atoms that are in the Selection's AtomGroup."
        return np.isin(self.universe.atoms.ix, self.ag.ix).astype(bool)

    def change_selection(
        self,
        selection_str: str,
        name: str,
        updating: bool = True,
        periodic: bool = True,
    ) -> None:
        "Change the current AtomGroup, using the parent universe and creating a new selection with the given `selectrion_str`"
        self.name = name
        self.periodic = periodic
        self.updating = updating
        self.selection_str = selection_str
        try:
            self.ag = self.universe.select_atoms(
                selection_str, updating=updating, periodic=periodic
            )
            self.message = ""
        except Exception as e:
            self.message = str(e)
            print(e)

    def to_mask(self) -> npt.NDArray[np.bool_]:
        "Returns the selection as a 1D numpy boolean mask. If updating=True, recomputes selection."
        if self.updating:
            self.mask_array = self._ag_to_mask()
        return self.mask_array

    @classmethod
    def from_atomgroup(cls, atomgroup: mda.AtomGroup, name: str = ""):
        "Create a Selection object from an AtomGroup"

        # set default value
        selection_str = f"sel_{atomgroup.n_atoms}_atoms"
        updating = False
        periodic = False

        # if class is an UpdatingAtomGroup
        if atomgroup.__class__.__name__ == "UpdatingAtomGroup":
            updating = True
            # assuming it's a single selection
            # MDA do support `u.select_atoms('index 0', 'around 5 index 0')
            selection_str = atomgroup._selection_strings[0]
            periodic = False
            try:
                if atomgroup._selections[0].periodic:
                    periodic = True
            except AttributeError as e:
                print(e)

        if name == "":
            name = selection_str
        selection = cls(atomgroup.universe, "all", name, updating, periodic)

        selection.selection_str = selection_str
        selection.ag = atomgroup
        selection.mask_array = selection._ag_to_mask()
        return selection



import numpy as np
import molecularnodes as mn
from molecularnodes.blender import mesh
import databpy
from databpy import BlenderObject
from databpy.object import LinkedObjectError
from .constants import data_dir
import bpy
import pytest


def test_bob():
    mol = mn.entities.fetch("8H1B", cache_dir=data_dir)
    assert isinstance(mol, BlenderObject)
    with pytest.raises(NotImplementedError):
        mol.set_frame(10)

    with pytest.raises(ValueError):
        mol.frames

    mol2 = mn.entities.fetch("1NMR", cache_dir=data_dir)
    assert isinstance(mol2.frames, bpy.types.Collection)
    assert mol2.name == "1NMR"


def test_set_position():
    mol = mn.entities.fetch("8FAT", cache_dir=data_dir)
    pos_a = mol.position
    mol.position += 10
    pos_b = mol.position
    assert not np.allclose(pos_a, pos_b)
    assert np.allclose(pos_a, pos_b - 10, rtol=0.1)


def test_eval_mesh():
    a = databpy.create_object(np.zeros((3, 3)))
    assert len(a.data.vertices) == 3
    b = databpy.create_object(np.zeros((5, 3)))
    assert len(b.data.vertices) == 5
    assert len(mesh.evaluate_using_mesh(b).data.vertices) == 5



import molecularnodes as mn
import bpy
import pytest
import numpy as np
import itertools
from .constants import data_dir
from .utils import NumpySnapshotExtension
from databpy import ObjectTracker

try:
    import pyopenvdb
except ImportError:
    pytest.skip("pyopenvdb not installed", allow_module_level=True)


@pytest.fixture
def density_file():
    file = data_dir / "emd_24805.map.gz"
    vdb_file = data_dir / "emd_24805.vdb"
    vdb_file.unlink(missing_ok=True)
    # Make all densities are removed
    for o in bpy.data.objects:
        if o.mn.entity_type == "density":
            bpy.data.objects.remove(o, do_unlink=True)
    return file


def test_density_load(density_file):
    density = mn.entities.density.load(density_file)
    pos = density.named_attribute("position")

    assert len(pos) > 1000

    avg = np.mean(pos, axis=0)
    assert np.linalg.norm(avg) > 0.5

    print(f"{list(bpy.data.objects)=}")

    assert density.object.mn.entity_type == "density"
    assert density.object.users_collection[0] == mn.blender.coll.mn()


def test_density_centered(density_file):
    # First load using standard parameters to test recreation of vdb
    density = mn.entities.density.load(density_file, center=True, overwrite=True)
    pos = density.named_attribute("position")
    avg = np.mean(pos, axis=0)
    assert len(pos) > 1000
    assert np.linalg.norm(avg) < 0.1


def test_density_invert(density_file):
    # First load using standard parameters to test recreation of vdb
    density = mn.entities.density.load(density_file)

    # Then refresh the scene
    bpy.data.objects.remove(density.object, do_unlink=True)

    density = mn.entities.density.load(density_file, invert=True)
    style_node = mn.blender.nodes.get_style_node(density.object)
    style_node.inputs["Threshold"].default_value = 0.01

    pos = density.named_attribute("position")
    # At this threshold after inverting we should have a cube the size of the volume
    assert pos[:, 0].max() > 2.0
    assert pos[:, 1].max() > 2.0
    assert pos[:, 2].max() > 2.0


def test_density_multiple_load():
    file = data_dir / "emd_24805.map.gz"
    density1 = mn.entities.density.load(file)
    density2 = mn.entities.density.load(file)

    assert density1.object.mn.entity_type == "density"
    assert density2.object.mn.entity_type == "density"
    assert density1.object.users_collection[0] == mn.blender.coll.mn()
    assert density2.object.users_collection[0] == mn.blender.coll.mn()


@pytest.mark.parametrize("name", ["", "NewDensity"])
def test_density_naming_op(density_file, name):
    bpy.context.scene.MN_import_density_name = name
    bpy.context.scene.MN_import_density = str(density_file)
    bpy.ops.mn.import_density()

    if name == "":
        object_name = "emd_24805"
    else:
        object_name = name
    object = bpy.data.objects[object_name]
    assert object
    assert object.name == object_name


@pytest.mark.parametrize("name", ["", "NewDensity"])
def test_density_naming_api(density_file, name):
    density = mn.entities.density.load(density_file, name)

    if name == "":
        object_name = "emd_24805"
    else:
        object_name = name

    assert density.object
    assert density.object.name == object_name


@pytest.mark.parametrize(
    "invert,node_setup,center", list(itertools.product([True, False], repeat=3))
)
def test_density_operator(
    snapshot_custom: NumpySnapshotExtension, density_file, invert, node_setup, center
):
    scene = bpy.context.scene
    scene.MN_import_density = str(density_file)
    scene.MN_import_density_invert = invert
    scene.MN_import_node_setup = node_setup
    scene.MN_import_density_center = center
    scene.MN_import_density_name = ""
    with ObjectTracker() as o:
        bpy.ops.mn.import_density()
        density = scene.MNSession.match(o.latest())

    assert snapshot_custom == density.position



from bpy.types import Material
from databpy.material import append_from_blend
from ..utils import MN_DATA_FILE

MATERIAL_NAMES = [
    "MN Default",
    "MN Flat Outline",
    "MN Squishy",
    "MN Transparent Outline",
    "MN Ambient Occlusion",
]


def append(name: str) -> Material:
    "Append a material from the MN_DATA_FILE."
    return append_from_blend(name, MN_DATA_FILE)


def add_all_materials() -> None:
    "Append all pre-defined materials from the MN_DATA_FILE."
    for name in MATERIAL_NAMES:
        append(name)


def default() -> Material:
    "Return the default material."
    return append("MN Default")



import bpy
from bpy.types import Context, Operator
from bpy.props import BoolProperty, EnumProperty, IntProperty, StringProperty

from ..blender import nodes
import databpy
from ..ui import node_info


def node_under_mouse(context, event):
    space = context.space_data
    mouse_pos = (event.mouse_region_x, event.mouse_region_y)

    # Find the node under the mouse
    node_under_mouse = None
    for node in space.node_tree.nodes:
        if (
            node.location.x < mouse_pos[0] < node.location.x + node.width
            and node.location.y < mouse_pos[1] < node.location.y + node.height
        ):
            node_under_mouse = node
            break

    return node_under_mouse


def _add_node(node_name, context, show_options=False, material="default"):
    """
    Add a node group to the node tree and set the values.

    intended to be called upon button press in the node tree, and not for use in general scripting
    """

    # actually invoke the operator to add a node to the current node tree
    # use_transform=True ensures it appears where the user's mouse is and is currently
    # being moved so the user can place it where they wish
    bpy.ops.node.add_node(
        "INVOKE_DEFAULT", type="GeometryNodeGroup", use_transform=True
    )
    node = context.active_node
    node.node_tree = bpy.data.node_groups[node_name]
    node.width = nodes.NODE_WIDTH
    node.show_options = show_options
    node.label = node_name
    node.name = node_name

    # if added node has a 'Material' input, set it to the default MN material
    nodes.assign_material(node, new_material=material)


class MN_OT_Add_Custom_Node_Group(Operator):
    bl_idname = "mn.add_custom_node_group"
    bl_label = "Add Custom Node Group"
    # bl_description = "Add Molecular Nodes custom node group."
    bl_options = {"REGISTER", "UNDO"}
    node_name: StringProperty(  # type: ignore
        name="node_name", description="", default="", subtype="NONE", maxlen=0
    )
    node_label: StringProperty(name="node_label", default="")  # type: ignore
    node_description: StringProperty(  # type: ignore
        name="node_description",
        description="",
        default="Add MolecularNodes custom node group.",
        subtype="NONE",
    )
    node_link: BoolProperty(name="node_link", default=True)  # type: ignore

    @classmethod
    def description(cls, context, properties):
        return properties.node_description

    def execute(self, context):
        try:
            nodes.append(self.node_name, link=self.node_link)
            _add_node(self.node_name, context)  # , label=self.node_label)
        except RuntimeError:
            self.report(
                {"ERROR"},
                message="Failed to add node. Ensure you are not in edit mode.",
            )
            return {"CANCELLED"}
        return {"FINISHED"}


class MN_OT_Assembly_Bio(Operator):
    bl_idname = "mn.assembly_bio"
    bl_label = "Build Biological Assembly"
    bl_description = "Adds node to build biological assembly based on symmetry operations that are extraced from the structure file"
    bl_options = {"REGISTER", "UNDO"}

    inset_node: BoolProperty(default=False)  # type: ignore

    @classmethod
    def poll(self, context):
        # this just checks to see that there is some biological assembly information that
        # is associated with the object / molecule. If there isn't then the assembly
        # operator will be greyed out and unable to be executed
        obj = context.active_object
        try:
            obj["biological_assemblies"]
            return True
        except KeyError:
            False

    def execute(self, context):
        obj = context.active_object
        with databpy.nodes.DuplicatePrevention():
            try:
                if self.inset_node:
                    nodes.assembly_insert(obj)
                else:
                    tree_assembly = nodes.assembly_initialise(obj)
                    _add_node(tree_assembly.name, context)
            except (KeyError, ValueError) as e:
                self.report({"ERROR"}, "Unable to build biological assembly node.")
                self.report({"ERROR"}, str(e))
                return {"CANCELLED"}

        return {"FINISHED"}


class MN_OT_iswitch_custom(Operator):
    bl_idname = "mn.iswitch_custom"
    # bl_idname = "mn.selection_custom"
    bl_label = "Custom ISwitch Node"
    bl_options = {"REGISTER", "UNDO"}

    description: StringProperty(name="Description")  # type: ignore
    dtype: EnumProperty(  # type: ignore
        name="Data type",
        items=(
            ("RGBA", "RGBA", "Color iswitch."),
            ("BOOLEAN", "BOOLEAN", "Boolean iswitch"),
        ),
    )
    field: StringProperty(name="field", default="chain_id")  # type: ignore
    prefix: StringProperty(name="prefix", default="Chain ")  # type: ignore
    node_property: StringProperty(name="node_property", default="chain_ids")  # type: ignore
    node_name: StringProperty(name="node_name", default="chain")  # type: ignore
    starting_value: IntProperty(name="starting_value", default=0)  # type: ignore

    @classmethod
    def description(cls, context, properties):
        return properties.description

    def execute(self, context):
        object = context.view_layer.objects.active
        prop = object[self.node_property]
        name = object.name
        if not prop:
            self.report(
                {"WARNING"},
                message=f"{self.node_property} not available for {object.name}.",
            )
            return {"CANCELLED"}

        prefix = {"BOOLEAN": "Select", "RGBA": "Color"}[self.dtype]
        node_name = " ".join([prefix, self.node_name, name])

        with databpy.nodes.DuplicatePrevention():
            node_chains = nodes.custom_iswitch(
                name=node_name,
                dtype=self.dtype,
                iter_list=prop,
                start=self.starting_value,
                field=self.field,
                prefix=self.prefix,
            )

        _add_node(node_chains.name, context)

        return {"FINISHED"}


class MN_OT_Residues_Selection_Custom(Operator):
    bl_idname = "mn.residues_selection_custom"
    bl_label = "Res ID Custom"
    bl_description = "Create a selection based on the provided residue strings.\nThis \
        node is built on a per-molecule basis, taking into account the residues that \
        were input."
    bl_options = {"REGISTER", "UNDO"}

    input_resid_string: StringProperty(  # type: ignore
        name="Select residue IDs: ",
        description="Enter a string value.",
        default="19,94,1-16",
    )

    def execute(self, context):
        with nodes.DuplicatePrevention():
            node_residues = nodes.resid_multiple_selection(
                node_name="MN_select_res_id_custom",
                input_resid_string=self.input_resid_string,
            )

        _add_node(node_residues.name, context)
        return {"FINISHED"}

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)


def get_swap_items(self, context):
    node = context.active_node
    prefix = node.node_tree.name.split(" ")[0].lower()
    if prefix == "is":
        prefix = "select"

    items = [
        (item.name, item.label, item.description)
        for item in node_info.menu_items.get_submenu(prefix).items
        if (not item.is_break and not item.is_custom and item.name != "Set Color")
    ]
    return items


class MN_OT_Node_Swap(Operator):
    bl_idname = "mn.node_swap"
    bl_label = "Swap Node"
    bl_description = "Swap this node for another."

    node_description: StringProperty(default="Swap selected node for another")  # type: ignore
    node_items: EnumProperty(items=get_swap_items)  # type: ignore

    @classmethod
    def description(cls, context, properties):
        return properties.node_description

    def execute(self, context: Context):
        node = context.active_node
        nodes.swap(node, self.node_items)
        return {"FINISHED"}


class MN_OT_Change_Color(Operator):
    bl_idname = "mn.change_color"
    bl_label = "Color"

    color: EnumProperty(  # type: ignore
        items=(
            (item.name, item.label, item.description)
            for item in node_info.menu_items.get_submenu("color").items
            if (not item.is_break and not item.is_custom and item.name != "Set Color")
        )
    )

    def execute(self, context: Context):
        node = context.active_node
        nodes.swap(node, self.color)
        self.report({"INFO"}, f"Selected {self.color}")
        return {"FINISHED"}


CLASSES = [
    MN_OT_Add_Custom_Node_Group,
    MN_OT_Residues_Selection_Custom,
    MN_OT_Assembly_Bio,
    MN_OT_iswitch_custom,
    MN_OT_Change_Color,
    MN_OT_Node_Swap,
]



import molecularnodes as mn
import pytest
import bpy
from .constants import data_dir


@pytest.mark.parametrize("format", ["bcif", "cif"])
def test_load_cellpack(snapshot, format):
    name = f"Cellpack_{format}"
    ens = mn.entities.ensemble.load_cellpack(
        data_dir / f"square1.{format}", name=name, node_setup=False, fraction=0.1
    )

    assert ens.name == name
    assert snapshot == str(ens.object["chain_ids"])
    obj_names = [obj.name for obj in ens.instance_collection.objects]
    assert snapshot == "\n".join(obj_names)

    ens.node_group.nodes["Ensemble Instance"].inputs["As Points"].default_value = False
    mn.blender.nodes.realize_instances(ens.object)
    for attribute in ens.list_attributes():
        assert snapshot == ens.named_attribute(attribute)

    pos_eval = ens.named_attribute("position", evaluate=True)
    assert snapshot == pos_eval.shape
    assert snapshot == pos_eval

    obj = ens.object



from pathlib import Path

import numpy as np
import bpy

from .ensemble import Ensemble
from .bcif import BCIF
from .cif import OldCIF
from ..molecule import molecule
from ... import blender as bl
from databpy import store_named_attribute, AttributeTypes
from ... import color


class CellPack(Ensemble):
    def __init__(self, file_path):
        super().__init__(file_path)
        self.file_type = self._file_type()
        self.chain_ids: np.ndarray
        self.transformations: np.ndarray
        self.array = self._read(self.file_path)

    def create_object(
        self,
        name="CellPack",
        node_setup: bool = True,
        world_scale: float = 0.01,
        fraction: float = 1.0,
        simplify=False,
    ):
        self.object = self._create_data_object(name=f"{name}")
        self._create_object_instances(name=name, node_setup=node_setup)
        self._setup_node_tree(fraction=fraction)
        return self.object

    def _file_type(self):
        return Path(self.file_path).suffix.strip(".")

    def _read(self, file_path):
        "Read a Cellpack File"
        suffix = Path(file_path).suffix

        if suffix in (".bin", ".bcif"):
            data = BCIF(file_path)
        elif suffix == ".cif":
            data = OldCIF(file_path)
        else:
            raise ValueError(f"Invalid file format: '{suffix}")

        self.chain_ids = data.chain_ids
        self.transformations = data.assemblies(as_array=True)
        return data.array

    def _create_object_instances(
        self, name: str = "CellPack", node_setup: bool = True
    ) -> bpy.types.Collection:
        collection = bl.coll.cellpack(name)

        if self.file_type == "cif":
            array = self.array[0]
        else:
            array = self.array
        for i, chain in enumerate(np.unique(array.chain_id)):
            chain_atoms = array[array.chain_id == chain]
            obj_name = f"{str(i).rjust(4, '0')}_{chain}"

            obj, coll_none = molecule._create_object(
                array=chain_atoms,
                name=obj_name,
                collection=collection,
            )

            colors = np.tile(color.random_rgb(i), (len(chain_atoms), 1))
            store_named_attribute(
                obj=obj,
                data=colors,
                name="Color",
                atype=AttributeTypes.FLOAT_COLOR,
            )

            if node_setup:
                bl.nodes.create_starting_node_tree(
                    obj, name=f"MN_pack_instance_{name}", color=None
                )

        self.data_collection = collection
        self.instance_collection = collection

        return collection

    def _create_data_object(self, name="DataObject"):
        data_object = bl.mesh.create_data_object(
            self.transformations, name=name, collection=bl.coll.mn()
        )

        data_object["chain_ids"] = self.chain_ids

        return data_object

    def _setup_node_tree(self, name="CellPack", fraction=1.0, as_points=False):
        mod = bl.nodes.get_mod(self.object)

        group = bl.nodes.new_tree(name=f"MN_ensemble_{name}", fallback=False)
        mod.node_group = group

        node_pack = bl.nodes.add_custom(group, "Ensemble Instance", location=[-100, 0])
        node_pack.inputs["Instances"].default_value = self.data_collection
        node_pack.inputs["Fraction"].default_value = fraction
        node_pack.inputs["As Points"].default_value = as_points

        link = group.links.new
        link(bl.nodes.get_input(group).outputs[0], node_pack.inputs[0])
        link(node_pack.outputs[0], bl.nodes.get_output(group).inputs[0])



"""
Importing molecular dynamics trajectories and associated files.
"""

__name__ = "MolecularNodes.trajectory"
__author__ = "Brady Johnston"

import bpy
import MDAnalysis as mda

from ... import blender as bl
from ...session import MNSession
from .trajectory import Trajectory
from . import dna
from bpy.props import StringProperty

bpy.types.Scene.MN_import_md_topology = StringProperty(
    name="Topology",
    description="File path for the toplogy file for the trajectory",
    subtype="FILE_PATH",
    maxlen=0,
)
bpy.types.Scene.MN_import_md_trajectory = StringProperty(
    name="Trajectory",
    description="File path for the trajectory file for the trajectory",
    subtype="FILE_PATH",
    maxlen=0,
)
bpy.types.Scene.MN_import_md_name = StringProperty(
    name="Name",
    description="Name of the molecule on import",
    default="NewTrajectory",
    maxlen=0,
)


def load(top, traj, name="NewTrajectory", style="spheres"):
    top = bl.path_resolve(top)
    traj = bl.path_resolve(traj)

    universe = mda.Universe(top, traj)

    traj = Trajectory(universe=universe)

    traj.create_object(name=name, style=style)

    return traj


class MN_OT_Reload_Trajectory(bpy.types.Operator):
    bl_idname = "mn.reload_trajectory"
    bl_label = "Reload Trajectory"
    bl_description = (
        "Reload the `mda.UNiverse` of the current Object to renable updating"
    )
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        traj = context.scene.MNSession.match(obj)
        return not traj

    def execute(self, context):
        obj = context.active_object
        session: MNSession = context.scene.MNSession
        topo = obj.mn.filepath_topology
        traj = obj.mn.filepath_trajectory

        if "oxdna" in obj.mn.entity_type:
            uni = mda.Universe(
                topo, traj, topology_format=dna.OXDNAParser, format=dna.OXDNAReader
            )
            traj = dna.OXDNA(uni)
        else:
            uni = mda.Universe(topo, traj)
            traj = Trajectory(uni)

        traj.object = obj
        traj.set_frame(context.scene.frame_current)
        return {"FINISHED"}


class MN_OT_Import_Trajectory(bpy.types.Operator):
    bl_idname = "mn.import_trajectory"
    bl_label = "Import Protein MD"
    bl_description = "Load molecular dynamics trajectory"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        scene = context.scene
        topology_file = scene.MN_import_md_topology
        trajectory_file = scene.MN_import_md_trajectory
        name = scene.MN_import_md_name

        trajectory = load(
            top=topology_file,
            traj=trajectory_file,
            name=name,
            style=scene.mn.import_style,
        )

        context.view_layer.objects.active = trajectory.object
        context.scene.frame_start = 0
        context.scene.frame_end = trajectory.universe.trajectory.n_frames

        self.report(
            {"INFO"},
            message=f"Imported '{topology_file}' as {trajectory.object.name} "
            f"with {str(trajectory.universe.trajectory.n_frames)} "
            f"frames from '{trajectory_file}'.",
        )

        return {"FINISHED"}


# UI


def panel(layout, scene):
    layout.label(text="Load MD Trajectories", icon="FILE_TICK")
    layout.separator()
    col = layout.column(align=True)
    row_import = col.row()
    row_import.prop(scene, "MN_import_md_name")
    row_import.operator("mn.import_trajectory", text="Load")
    col.separator()
    col.prop(scene, "MN_import_md_topology")
    col.prop(scene, "MN_import_md_trajectory")

    layout.separator()
    layout.label(text="Options", icon="MODIFIER")
    row = layout.row()
    row.prop(scene, "MN_import_node_setup", text="")
    col = row.column()
    col.prop(scene.mn, "import_style")
    col.enabled = scene.MN_import_node_setup


CLASSES = [MN_OT_Import_Trajectory, MN_OT_Reload_Trajectory]



from pathlib import Path

import bpy
from bpy.types import Context, UILayout
from bpy.props import EnumProperty, StringProperty, BoolProperty, CollectionProperty
from biotite import InvalidFileError
import os
import io

from ...download import FileDownloadPDBError, download, CACHE_DIR
from ...blender import path_resolve
from ..ensemble.cif import OldCIF
from .molecule import Molecule
from .pdb import PDB
from .pdbx import BCIF, CIF
from .sdf import SDF
from ...style import STYLE_ITEMS


def parse(filepath) -> Molecule:
    # TODO: I don't like that we might be dealing with bytes or a filepath here,
    # I need to work out a nicer way to have it be cleanly one or the other

    if isinstance(filepath, io.BytesIO):
        suffix = ".bcif"
    else:
        filepath = path_resolve(filepath)
        suffix = Path(filepath).suffix

    parser = {
        ".pdb": PDB,
        ".pdbx": CIF,
        ".cif": CIF,
        ".bcif": BCIF,
        ".mol": SDF,
        ".sdf": SDF,
    }

    if suffix not in parser:
        raise ValueError(f"Unable to open local file. Format '{suffix}' not supported.")
    try:
        molecule = parser[suffix](filepath)
    except InvalidFileError:
        molecule = OldCIF(filepath)

    return molecule


def fetch(
    pdb_code: str,
    style: str | None = "spheres",
    centre: str = "",
    del_solvent: bool = True,
    del_hydrogen: bool = False,
    cache_dir: str | None = None,
    build_assembly: bool = False,
    database: str = "rcsb",
    format: str = "bcif",
    color: str = "common",
) -> Molecule:
    if build_assembly:
        centre = ""

    file_path = download(
        code=pdb_code, format=format, cache=cache_dir, database=database
    )

    mol = parse(file_path)

    obj = mol.create_object(
        name=pdb_code,
        centre=centre,
        style=style,
        del_solvent=del_solvent,
        del_hydrogen=del_hydrogen,
        build_assembly=build_assembly,
        color=color,
    )

    obj.mn["pdb_code"] = pdb_code
    obj.mn["entity_type"] = format

    return mol


def load_local(
    file_path,
    name="Name",
    centre="",
    del_solvent=True,
    del_hydrogen=False,
    style="spheres",
    build_assembly=False,
):
    mol = parse(file_path)
    mol.create_object(
        name=name,
        style=style,
        build_assembly=build_assembly,
        centre=centre,
        del_solvent=del_solvent,
        del_hydrogen=del_hydrogen,
    )
    return mol


class Import_Molecule(bpy.types.Operator):
    style: EnumProperty(  # type: ignore
        name="Style",
        default="spheres",
        description="Starting style for the structure on import",
        items=STYLE_ITEMS,
    )

    node_setup: BoolProperty(  # type: ignore
        name="Node Setup",
        default=True,
        description="Whether to setup the starting default node tree on import",
    )
    centre: EnumProperty(  # type: ignore
        name="Centre",
        description="Centre the structure at the world origin using the given method",
        default="None",
        items=(
            ("None", "None", "No centering is applied", 1),
            (
                "mass",
                "Mass",
                "Adjust the structure's centre of mass to be at the world origin",
                2,
            ),
            (
                "centroid",
                "Centroid",
                "Adjust the structure's centroid (centre of geometry) to be at the world origin",
                3,
            ),
        ),
    )
    del_solvent: BoolProperty(  # type: ignore
        default=True,
        name="Delete Solvent",
        description="Remove solvent atoms from the structure on import",
    )
    assembly: BoolProperty(  # type: ignore
        default=False,
        name="Build Biological Assembly",
        description="Build the biological assembly for the structure on import",
    )

    def draw(self, context: Context) -> UILayout:
        layout = self.layout
        row = layout.row()
        row.prop(self, "node_setup", text="")
        col = row.column()
        col.prop(self, "style")
        col.enabled = self.node_setup
        # row = layout.row()
        layout.prop(self, "centre")
        layout.prop(self, "del_solvent")
        layout.prop(self, "assembly")

        return layout


class MN_OT_Import_Molecule(Import_Molecule):
    bl_idname = "mn.import_molecule"
    bl_label = "Import a Molecule"

    directory: StringProperty(  # type: ignore
        subtype="FILE_PATH", options={"SKIP_SAVE", "HIDDEN"}
    )
    files: CollectionProperty(  # type: ignore
        type=bpy.types.OperatorFileListElement, options={"SKIP_SAVE", "HIDDEN"}
    )

    def draw(self, context: Context):
        layout = self.layout
        layout.label(text=f"Importing {len(self.files)} molecules")
        layout = super().draw(context)

    def execute(self, context):
        if not self.directory:
            return {"CANCELLED"}

        if not self.node_setup:
            style = None
        else:
            style = self.style

        for file in self.files:
            try:
                mol = parse(os.path.join(self.directory, file.name))
                mol.create_object(
                    name=file.name,
                    centre=self.centre,
                    style=style,
                    del_solvent=self.del_solvent,
                    build_assembly=self.assembly,
                )
            except Exception as e:
                print(f"Failed importing {file}: {e}")

        return {"FINISHED"}

    def invoke(self, context, event):
        if context.area and context.area.type == "VIEW_3D":
            context.window_manager.invoke_props_dialog(self)
        else:
            context.window_manager.fileselect_add(self)
        return {"RUNNING_MODAL"}


class MN_FH_Import_Molecule(bpy.types.FileHandler):
    bl_idname = "MN_FH_import_molecule"
    bl_label = "File handler for import molecular data files."
    bl_import_operator = "mn.import_molecule"
    bl_file_extensions = ".pdb;.cif;.mmcif;.bcif;.pdbx"

    @classmethod
    def poll_drop(cls, context):
        return context.area and context.area.type == "VIEW_3D"


DOWNLOAD_FORMATS = (
    ("bcif", ".bcif", "Binary compressed .cif file, fastest for downloading"),
    ("cif", ".cif", "The new standard of .cif / .mmcif"),
    ("pdb", ".pdb", "The classic (and depcrecated) PDB format"),
)


class MN_OT_Import_Fetch(Import_Molecule):
    bl_idname = "mn.import_fetch"
    bl_label = "Download a Molecule"
    bl_description = "Download a molecule from the wwPDB and import it to the scene"

    code: StringProperty(  # type: ignore
        default="4ozs",
        name="PDB Code",
        description="Code to use for downloading from the wwPDB",
    )
    file_format: EnumProperty(  # type: ignore
        name="Format",
        description="Format to download as from the PDB",
        default="bcif",
        items=DOWNLOAD_FORMATS,
    )

    database: EnumProperty(  # type: ignore
        name="Database",
        description="Database to download structure from",
        default="rcsb",
        items=(
            ("rcsb", "rcsb", "The RCSB PDB"),
            ("pdbe", "PDBe", "The european protein data bank"),
            ("pdbj", "PDBj", "The Japanese protein data bank"),
            ("alphafold", "AplhaFold", "The AlphaFold database"),
            ("emdb", "EM Database", "Electron microscopy database"),
        ),
    )

    def execute(self, context):
        try:
            file_path = download(
                self.code, format=self.file_format, cache=self.cache, database="rcsb"
            )
            mol = parse(file_path)
            mol.create_object(
                name=self.code,
                style=self.style,
                centre=self.centre,
                del_solvent=self.del_solvent,
                build_assembly=self.assembly,
            )
        except Exception:
            return {"CANCELLED"}
        return {"FINISHED"}

    def invoke(self, context, event):
        context.window_manager.invoke_props_dialog(self)
        return {"RUNNING_MODAL"}


# Properties that can be set in the scene, to be passed to the operator


bpy.types.Scene.MN_pdb_code = StringProperty(
    name="PDB",
    description="The 4-character PDB code to download",
    options={"TEXTEDIT_UPDATE"},
    maxlen=4,
)
bpy.types.Scene.MN_cache_dir = StringProperty(
    name="",
    description="Directory to save the downloaded files",
    options={"TEXTEDIT_UPDATE"},
    default=CACHE_DIR,
    subtype="DIR_PATH",
)
bpy.types.Scene.MN_cache = BoolProperty(
    name="Cache Downloads",
    description="Save the downloaded file in the given directory",
    default=True,
)
bpy.types.Scene.MN_import_del_hydrogen = BoolProperty(
    name="Remove Hydrogens",
    description="Remove the hydrogens from a structure on import",
    default=False,
)
bpy.types.Scene.MN_import_format_download = EnumProperty(
    name="Format",
    description="Format to download as from the PDB",
    items=(
        ("bcif", ".bcif", "Binary compressed .cif file, fastest for downloading"),
        ("cif", ".cif", "The new standard of .cif / .mmcif"),
        ("pdb", ".pdb", "The classic (and depcrecated) PDB format"),
    ),
)
bpy.types.Scene.MN_import_local_path = StringProperty(
    name="File",
    description="File path of the structure to open",
    options={"TEXTEDIT_UPDATE"},
    subtype="FILE_PATH",
    maxlen=0,
)
bpy.types.Scene.MN_import_local_name = StringProperty(
    name="Name",
    description="Name of the molecule on import",
    options={"TEXTEDIT_UPDATE"},
    default="NewMolecule",
    maxlen=0,
)

bpy.types.Scene.MN_alphafold_code = StringProperty(
    name="UniProt ID",
    description="The UniProt ID to use for downloading from the AlphaFold databse",
    options={"TEXTEDIT_UPDATE"},
)

bpy.types.Scene.MN_import_format_alphafold = EnumProperty(
    name="Format",
    description="Format to download as from the PDB",
    items=(
        # ("bcif", ".bcif", "Binary compressed .cif file, fastest for downloading"),
        ("cif", ".cif", "The new standard of .cif / .mmcif"),
        ("pdb", ".pdb", "The classic (and depcrecated) PDB format"),
    ),
)


# operator that is called by the 'button' press which calls the fetch function


class MN_OT_Import_wwPDB(bpy.types.Operator):
    bl_idname = "mn.import_wwpdb"
    bl_label = "Fetch"
    bl_description = "Download and open a structure from the Protein Data Bank"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        scene = context.scene
        pdb_code = scene.MN_pdb_code
        cache_dir = scene.MN_cache_dir
        file_format = scene.MN_import_format_download

        if not scene.MN_cache:
            cache_dir = None

        style = None
        if scene.MN_import_node_setup:
            style = scene.mn.import_style

        centre = ""
        if scene.MN_import_centre:
            centre = scene.MN_centre_type

        try:
            mol = fetch(
                pdb_code=pdb_code,
                centre=centre,
                del_solvent=scene.MN_import_del_solvent,
                del_hydrogen=scene.MN_import_del_hydrogen,
                style=style,
                cache_dir=cache_dir,
                build_assembly=scene.MN_import_build_assembly,
                format=file_format,
            )
        except FileDownloadPDBError as e:
            self.report({"ERROR"}, str(e))
            if file_format == "pdb":
                self.report(
                    {"ERROR"},
                    "There may not be a `.pdb` formatted file available - try a different download format.",
                )
            return {"CANCELLED"}

        bpy.context.view_layer.objects.active = mol.object
        self.report({"INFO"}, message=f"Imported '{pdb_code}' as {mol.object.name}")

        return {"FINISHED"}


class MN_OT_Import_Protein_Local(bpy.types.Operator):
    bl_idname = "mn.import_protein_local"
    bl_label = "Load"
    bl_description = "Open a local structure file"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        scene = context.scene
        file_path = scene.MN_import_local_path

        style = scene.mn.import_style
        if not scene.MN_import_node_setup:
            style = None

        centre = ""
        if scene.MN_import_centre:
            centre = scene.MN_centre_type

        mol = load_local(
            file_path=file_path,
            name=scene.MN_import_local_name,
            centre=centre,
            del_solvent=scene.MN_import_del_solvent,
            del_hydrogen=scene.MN_import_del_hydrogen,
            style=style,
            build_assembly=scene.MN_import_build_assembly,
        )

        # return the good news!
        bpy.context.view_layer.objects.active = mol.object
        self.report({"INFO"}, message=f"Imported '{file_path}' as {mol.name}")
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)


class MN_OT_Import_AlphaFold(bpy.types.Operator):
    bl_idname = "mn.import_alphafold"
    bl_label = "Fetch"
    bl_description = "Download specified structure from the AlphaFold databse"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        scene = context.scene
        pdb_code = scene.MN_alphafold_code.strip()
        cache_dir = scene.MN_cache_dir
        file_format = scene.MN_import_format_alphafold

        if not scene.MN_cache:
            cache_dir = None

        style = None
        if scene.MN_import_node_setup:
            style = scene.mn.import_style

        centre = ""
        if scene.MN_import_centre:
            centre = scene.MN_centre_type

        try:
            mol = fetch(
                pdb_code=pdb_code,
                centre=centre,
                del_solvent=scene.MN_import_del_solvent,
                style=style,
                cache_dir=cache_dir,
                build_assembly=scene.MN_import_build_assembly,
                format=file_format,
                database="alphafold",
                color="plddt",
            )
        except FileDownloadPDBError as e:
            self.report({"ERROR"}, str(e))
            if file_format == "pdb":
                self.report(
                    {"ERROR"},
                    "There may not be a `.pdb` formatted file available - try a different download format.",
                )
            return {"CANCELLED"}

        bpy.context.view_layer.objects.active = mol.object
        self.report({"INFO"}, message=f"Imported '{pdb_code}' as {mol.object.name}")

        return {"FINISHED"}


# the UI for the panel, which will display the operator and the properties


def check_online_access_for_ui(layout: bpy.types.UILayout) -> bpy.types.UILayout:
    """
    Disable UI without Online Access

    Checks for the online access permissions, and adds a warning and disables following
    UI elements if it fails the check. Returns the UILayout that will have .enabled flag
    set to False, disabling all subsequent uses of the layout.

    Args:
        layout (bpy.types.UILayout): The UILayout element to add the warning and potentially
        disable.

    Returns:
        bpy.types.UILayout: The altered UILayout element, for use in downstream UI
        components.
    """
    if not bpy.app.online_access:
        layout.label(
            text="Online access disabled. Change in Blender's system preferences.",
            icon="ERROR",
        )
        op = layout.operator("wm.url_open", text="Online Access Docs", icon="URL")
        op.url = "https://docs.blender.org/manual/en/dev/editors/preferences/system.html#bpy-types-preferencessystem-use-online-access"
        layout = layout.column()
        layout.alert = True
        layout.enabled = False

    return layout


def panel_wwpdb(layout, scene):
    layout.label(text="Download from PDB", icon="IMPORT")
    layout.separator()

    layout = check_online_access_for_ui(layout)

    row_import = layout.row().split(factor=0.5)
    row_import.prop(scene, "MN_pdb_code")
    download = row_import.split(factor=0.3)
    download.prop(scene, "MN_import_format_download", text="")
    download.operator("mn.import_wwpdb")
    layout.separator(factor=0.4)

    row = layout.row().split(factor=0.3)
    row.prop(scene, "MN_cache")
    row_cache = row.row()
    row_cache.prop(scene, "MN_cache_dir")
    row_cache.enabled = scene.MN_cache
    layout.separator()

    layout.label(text="Options", icon="MODIFIER")
    options = layout.column(align=True)

    row = options.row()
    row.prop(scene, "MN_import_node_setup", text="")
    col = row.column()
    col.prop(scene.mn, "import_style")
    col.enabled = scene.MN_import_node_setup

    row_centre = options.row()
    row_centre.prop(scene, "MN_import_centre", icon_value=0)
    col_centre = row_centre.column()
    col_centre.prop(scene, "MN_centre_type", text="")
    col_centre.enabled = scene.MN_import_centre
    options.separator()

    grid = options.grid_flow()
    grid.prop(scene, "MN_import_build_assembly")
    grid.prop(scene, "MN_import_del_solvent")
    grid.prop(scene, "MN_import_del_hydrogen")


def panel_alphafold(layout, scene):
    layout.label(text="Download from the AlphaFold DataBase", icon="IMPORT")
    layout.separator()

    layout = check_online_access_for_ui(layout)

    row_import = layout.row().split(factor=0.5)
    row_import.prop(scene, "MN_alphafold_code")
    download = row_import.split(factor=0.3)
    download.prop(scene, "MN_import_format_alphafold", text="")
    download.operator("mn.import_alphafold")
    layout.separator(factor=0.4)

    row = layout.row().split(factor=0.3)
    row.prop(scene, "MN_cache")
    row_cache = row.row()
    row_cache.prop(scene, "MN_cache_dir")
    row_cache.enabled = scene.MN_cache
    layout.separator()

    layout.label(text="Options", icon="MODIFIER")
    options = layout.column(align=True)

    row = options.row()
    row.prop(scene, "MN_import_node_setup", text="")
    col = row.column()
    col.prop(scene.mn, "import_style")
    col.enabled = scene.MN_import_node_setup

    row_centre = options.row()
    row_centre.prop(scene, "MN_import_centre", icon_value=0)
    col_centre = row_centre.column()
    col_centre.prop(scene, "MN_centre_type", text="")
    col_centre.enabled = scene.MN_import_centre
    options.separator()

    grid = options.grid_flow()
    grid.prop(scene, "MN_import_build_assembly")
    grid.prop(scene, "MN_import_del_solvent")
    # grid.prop(scene, "MN_import_del_hydrogen")


# operator that calls the function to import the structure from a local file


def panel_local(layout, scene):
    layout.label(text="Load a Local File", icon="FILE_TICK")
    layout.separator()

    row_name = layout.row(align=False)
    row_name.prop(scene, "MN_import_local_name")
    row_name.operator("mn.import_protein_local")

    row_import = layout.row()
    row_import.prop(scene, "MN_import_local_path")
    layout.separator()

    layout.label(text="Options", icon="MODIFIER")
    options = layout.column(align=True)

    row = options.row()
    row.prop(scene, "MN_import_node_setup", text="")
    col = row.column()
    col.prop(scene.mn, "import_style")
    col.enabled = scene.MN_import_node_setup

    row_centre = options.row()

    row_centre.prop(scene, "MN_import_centre", icon_value=0)
    # row_centre.prop()
    col_centre = row_centre.column()
    col_centre.prop(scene, "MN_centre_type", text="")
    col_centre.enabled = scene.MN_import_centre
    options.separator()

    grid = options.grid_flow()
    grid.prop(scene, "MN_import_build_assembly")
    grid.prop(scene, "MN_import_del_solvent", icon_value=0)
    grid.prop(scene, "MN_import_del_hydrogen", icon_value=0)


CLASSES = [
    MN_OT_Import_AlphaFold,
    MN_OT_Import_Protein_Local,
    MN_OT_Import_wwPDB,
    MN_OT_Import_Molecule,
    MN_FH_Import_Molecule,
    MN_OT_Import_Fetch,
]



from typing import List, Union
import bpy


class Item:
    def __init__(self) -> None:
        self.is_break = False
        self.is_custom = False
        self.backup: str = None

    @classmethod
    def menu(
        self, layout: bpy.types.UILayout, context: bpy.types.Context = None
    ) -> None:
        pass

    @property
    def node_name(self) -> str:
        if hasattr(self, "backup"):
            return self.backup
        else:
            return self.name

    @property
    def tree(self) -> bpy.types.GeometryNodeGroup:
        if self.name.startswith("mn."):
            return bpy.data.node_groups[self.backup]
        return bpy.data.node_groups[self.name]


class MenuItem(Item):
    def __init__(
        self,
        name: str,
        label: str | None = None,
        description: str = "",
        videos: list[str] | None = None,
        backup: str | None = None,
    ) -> None:
        super().__init__()
        self.name = name
        self.label = label
        if self.label is None:
            self.label = self.name
        self.description = description
        self._videos = videos
        self.backup = backup

    def short_description(self):
        return self.description.split("\n")[0].removesuffix(".")

    @property
    def videos(self) -> list[str]:
        if self._videos is None:
            return []
        if isinstance(self._videos, str):
            return [self._videos]
        return self._videos

    def menu(
        self, layout: bpy.types.UILayout, context: bpy.types.Context = None
    ) -> None:
        if self.label is None:
            self.label = self.name

        if self.name.startswith("mn."):
            layout.operator(self.name)
            return None

        op = layout.operator("mn.add_custom_node_group", text=self.label)
        op.node_label = self.label
        op.node_name = self.name
        op.node_description = self.description
        op.node_link = False

    def to_dict(self) -> dict:
        "Return a dictionary with key: node_name and value: node_data (name, label, description, links, backup)"
        return {
            self.name: {
                "name": self.name,
                "label": self.label,
                "description": self.description,
                "videos": self.videos,
                "backup": self.backup,
            }
        }


class CustomItem(Item):
    def __init__(
        self,
        label: str,
        field: str,
        dtype: str,
        name: str,
        prefix: str,
        property_id: str,
        description: str,
        videos: str = None,
    ) -> None:
        super().__init__()
        self.label = label
        self.field = field
        self.dtype = dtype
        self.name = name
        self.prefix = prefix
        self.property_id = property_id
        self.description = description
        self.videos = videos
        self.is_custom = True

    def menu(
        self, layout: bpy.types.UILayout, context: bpy.types.Context = None
    ) -> None:
        row = layout.row()
        op = row.operator("mn.iswitch_custom", text=self.label)
        op.field = self.field
        op.dtype = self.dtype
        op.prefix = self.prefix
        op.node_property = self.property_id
        op.node_name = self.label

        if self.dtype == "RGBA":
            op.description = f"Choose custom colors for {self.label}"
        elif self.dtype == "BOOLEAN":
            op.description = f"Choose custom selections for {self.label}"
        else:
            raise ValueError(f"Data type currently not supported: {self.dtype}")
        # test if the object has the currently tested property to enable operator
        row.enabled = bool(context.active_object.get(self.property_id))


class Break:
    def __init__(self, text: str = None) -> None:
        super().__init__()
        self.is_break = True
        self.text = text

    def menu(
        self, layout: bpy.types.UILayout, context: bpy.types.Context = None
    ) -> None:
        layout.separator()
        # optionally we can add a subtitle for the next section of nodes
        if self.text and self.text.strip() != "":
            layout.label(text=self.text)


class Submenu:
    def __init__(self, name, items, title: str = None, description: str = None) -> None:
        self.name: str = name
        self.items: List[Union[MenuItem, Break, CustomItem]] = items
        self.title = title
        self.description = description

    def node_names(self):
        return [item.name for item in self.items if not isinstance(item, Break)]

    def menu(self, layout: bpy.types.UILayout, context: bpy.types.Context):
        for item in self.items:
            item.menu(layout=layout, context=context)


class Menu:
    def __init__(self, submenus: List[Submenu]) -> None:
        self.submenus = submenus

    def get_submenu(self, name: str) -> Submenu:
        for sub in self.submenus:
            if sub.name == name:
                return sub



import pytest
import molecularnodes as mn

from .constants import data_dir


@pytest.fixture
def filepath():
    return data_dir / "1f2n.bcif"


def test_bcif_init(filepath):
    bcif = mn.entities.BCIF(filepath)
    assert bcif.file_path == filepath


def test_bcif_read(filepath):
    bcif = mn.entities.BCIF(filepath)
    assert bcif.file is not None


def test_bcif_get_structure(filepath):
    bcif = mn.entities.BCIF(filepath)
    structure = bcif.get_structure()
    assert structure is not None
    # assert structure.shape[1] == len(bcif)


def test_bcif_assemblies(filepath):
    bcif = mn.entities.BCIF(filepath)
    assemblies = bcif.assemblies()
    assert assemblies is not None


def test_bcif_entity_ids(filepath):
    bcif = mn.entities.BCIF(filepath)
    entity_ids = bcif.entity_ids
    assert entity_ids is not None
    assert entity_ids == ["CAPSID PROTEIN", "CALCIUM ION", "water"]



import numpy as np

from MDAnalysis.coordinates.base import ReaderBase
from MDAnalysis.lib import util


def _is_info_line(line: str):
    return line.startswith("t = ") or line.startswith("b = ") or line.startswith("E = ")


class OXDNAReader(ReaderBase):
    def __init__(self, filename, **kwargs):
        super(OXDNAReader, self).__init__(filename, **kwargs)

        self.n_atoms = kwargs["n_atoms"]
        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)

        oxdnafile = self._oxdnafile = util.anyopen(filename, "rb")

        starts = []
        stops = []

        line = "a new line"
        previous_is_info = False
        current_is_info = False
        while line:
            line = oxdnafile.readline().decode()
            current_is_info = _is_info_line(line)

            # get the lines where the data values start and stop, which
            # are separated by the several lines of overall information
            if not current_is_info and previous_is_info:
                starts.append(oxdnafile.tell() - len(line))
            if current_is_info and not previous_is_info:
                stops.append(oxdnafile.tell() - len(line))

            previous_is_info = current_is_info

        # add the end of the file
        stops += [oxdnafile.tell()]

        self._start_offsets = starts
        self._stop_offsets = stops[1:]  # drop the first
        self.n_frames = len(self._start_offsets)

        self._read_frame(0)

    def _reopen(self):
        self.close()
        self._oxdnafile = util.anyopen(self.filename, "rb")
        self.ts.frame = -1

    def _read_next_timestep(self, ts=None):
        frame = self.frame + 1
        return self._read_frame(frame)

    def _read_frame(self, frame):
        try:
            start = self._start_offsets[frame]
            stop = self._stop_offsets[frame]
        except IndexError:
            raise OSError from None

        self._oxdnafile.seek(start)
        chunk = self._oxdnafile.read(stop - start)

        array = np.array(
            [
                np.array(line.split(), dtype=float)
                for line in chunk.decode().splitlines()
            ]
        )

        # TODO: also access and update the other values
        self.ts.positions = array[:, :3]

        for i, name in enumerate(
            ("base_vector", "base_normal", "velocity", "angular_velocity")
        ):
            starting_column = 3 * (i + 1)
            if starting_column >= array.shape[1]:
                continue
            cols = np.arange(3) + starting_column
            self.ts.data[name] = array[:, cols]
        self.ts.frame = frame

        return self.ts

    def close(self):
        self._oxdnafile.close()



import bpy
import pytest
import numpy as np
import molecularnodes as mn

from databpy import ObjectTracker

from .utils import NumpySnapshotExtension
from .constants import data_dir, codes, attributes

# register the operators, which isn't done by default when loading bpy
# just via headless float_decimals


@pytest.mark.parametrize("code", codes)
def test_op_api_cartoon(
    snapshot_custom: NumpySnapshotExtension, code, style="ribbon", format="bcif"
):
    scene = bpy.context.scene
    scene.MN_import_node_setup = True
    scene.MN_pdb_code = code
    scene.mn.import_style = style
    scene.MN_import_node_setup = True
    scene.MN_import_build_assembly = False
    scene.MN_import_centre = False
    scene.MN_import_del_solvent = False
    scene.MN_import_format_download = format

    bpy.ops.mn.import_wwpdb()
    mol1 = scene.MNSession.match(bpy.context.active_object)

    mol2 = mn.entities.fetch(code, style=style, format=format, cache_dir=data_dir)

    # objects being imported via each method should have identical snapshots
    for mol in [mol1, mol2]:
        for name in attributes:
            try:
                assert snapshot_custom == mol.named_attribute(
                    "position", evaluate=False
                )
            except AttributeError as e:
                assert snapshot_custom == str(e)


@pytest.mark.parametrize("code", codes)
@pytest.mark.parametrize("file_format", ["bcif", "cif", "pdb"])
def test_op_local(snapshot_custom, code, file_format):
    scene = bpy.context.scene
    session = scene.MNSession
    scene.MN_import_node_setup = False
    scene.mn.import_style = "spheres"
    scene.MN_import_build_assembly = False
    scene.MN_import_del_solvent = False
    scene.MN_import_format_download = file_format
    path = str(mn.download.download(code=code, format=file_format, cache=data_dir))
    scene.MN_import_local_path = path
    scene.MN_centre_type = "centroid"

    scene.MN_import_centre = False
    with ObjectTracker() as o:
        bpy.ops.mn.import_protein_local()
        mol = session.match(o.latest())

    scene.MN_import_centre = True
    with ObjectTracker() as o:
        bpy.ops.mn.import_protein_local()
        mol_cent = session.match(o.latest())

    assert snapshot_custom == mol.position
    assert snapshot_custom == mol_cent.position
    assert not np.allclose(mol.position, mol_cent.position)


def test_op_api_mda(snapshot_custom: NumpySnapshotExtension):
    bpy.context.scene.frame_set(0)

    topo = str(data_dir / "md_ppr/box.gro")
    traj = str(data_dir / "md_ppr/first_5_frames.xtc")
    name = bpy.context.scene.MN_import_md_name

    bpy.context.scene.MN_import_md_topology = topo
    bpy.context.scene.MN_import_md_trajectory = traj
    bpy.context.scene.mn.import_style = "ribbon"

    with ObjectTracker() as o:
        bpy.ops.mn.import_trajectory()
        obj_1 = o.latest()

    traj_op = bpy.context.scene.MNSession.match(obj_1)
    assert traj_op.name == name

    traj_func = mn.entities.trajectory.load(topo, traj, name="test", style="ribbon")

    bpy.context.scene.frame_set(2)
    assert np.allclose(traj_func.position, traj_op.position)
    pos_2 = traj_func.position.copy()
    bpy.context.scene.frame_set(4)
    traj_op.set_frame(4)
    traj_func.set_frame(4)

    assert not np.allclose(pos_2, traj_op.position)
    assert not np.allclose(pos_2, traj_func.position)



import bpy
from bpy.app.handlers import persistent
# from .session import update_trajectories

# from .session import MNSession


# this update function requires a self and context input, as funcitons with these inputs
# have ot be passed to the `update` arguments of UI properties. When the UI is updated,
# the function is called with the UI element being `self` and the current context being
# passed into the function
def _update_trajectories(self, context: bpy.types.Context) -> None:
    """
    Function for being called at various points in the updating of the UI, to ensure
    positions and selections of the trajectories are udpated with the new inputs
    """
    update_trajectories(context.scene)


def _update_trajectories_on_frame_change(self, context: bpy.types.Context) -> None:
    """
    Function for being called at various points in the updating of the UI, to ensure
    positions and selections of the trajectories are udpated with the new inputs
    """
    update_trajectories(context.scene)


def _selection_update_trajectories(self, context: bpy.types.Context) -> None:
    """
    Function for selection changing. If the selection is immutable e.g.
    when it is generated from an AtomGroup,
    update universe will not be called to avoid invalid selection str.
    """
    if self.immutable:
        return
    else:
        update_trajectories(context.scene)


# this is the 'perisisent' function which can be appended onto the
# `bpy.app.handlers.frame_change_*` functions. Either before or after the frame changes
# this function will then be called - ensuring all of the trajectories are up to date. We
# use the `frame_change_post` handler as we want the frame to change first, then we update
# the universe based on the current frame value
@persistent
def update_trajectories(scene):
    "Call the set_frame method of all trajectories in the current session"
    session = scene.MNSession
    for traj in session.trajectories.values():
        # use the updated method if it exists but otherwise fallback on the old method
        # of updating the trajectories
        if hasattr(traj, "update_with_scene"):
            if traj.update_with_scene:
                traj.set_frame(scene.frame_current)
            else:
                traj.set_frame(traj.frame)

        else:
            traj._update_positions(scene.frame_current)
            traj._update_selections()
            traj._update_calculations()

        # except NotImplementedError:
        #     pass
        # except Exception as e:
        #     # print(f"Error updating {traj}: {e}")
        #     raise e



import bpy
from .star import StarFile

from .cellpack import CellPack

bpy.types.Scene.MN_import_star_file_path = bpy.props.StringProperty(
    name="File",
    description="File path for the `.star` file to import.",
    subtype="FILE_PATH",
    maxlen=0,
)
bpy.types.Scene.MN_import_star_file_name = bpy.props.StringProperty(
    name="Name",
    description="Name of the created object.",
    default="NewStarInstances",
    maxlen=0,
)


def load_starfile(
    file_path, name="NewStarInstances", node_setup=True, world_scale=0.01
):
    ensemble = StarFile.from_starfile(file_path)
    ensemble.create_object(name=name, node_setup=node_setup, world_scale=world_scale)

    return ensemble


class MN_OT_Import_Star_File(bpy.types.Operator):
    bl_idname = "mn.import_star_file"
    bl_label = "Load"
    bl_description = (
        "Will import the given file, setting up the points to instance an object."
    )
    bl_options = {"REGISTER"}

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        scene = context.scene
        load_starfile(
            file_path=scene.MN_import_star_file_path,
            name=scene.MN_import_star_file_name,
            node_setup=True,
        )
        return {"FINISHED"}


def panel_starfile(layout, scene):
    layout.label(text="Load Star File", icon="FILE_TICK")
    layout.separator()
    row_import = layout.row()
    row_import.prop(scene, "MN_import_star_file_name")
    layout.prop(scene, "MN_import_star_file_path")
    row_import.operator("mn.import_star_file")


bpy.types.Scene.mol_import_cell_pack_path = bpy.props.StringProperty(
    name="File",
    description="File to import (.cif, .bcif)",
    subtype="FILE_PATH",
    maxlen=0,
)
bpy.types.Scene.mol_import_cell_pack_name = bpy.props.StringProperty(
    name="Name",
    description="Name of the created object.",
    default="NewCellPackModel",
    maxlen=0,
)


def load_cellpack(
    file_path,
    name="NewCellPackModel",
    node_setup=True,
    world_scale=0.01,
    fraction: float = 1,
):
    ensemble = CellPack(file_path)
    ensemble.create_object(
        name=name, node_setup=node_setup, world_scale=world_scale, fraction=fraction
    )

    return ensemble


class MN_OT_Import_Cell_Pack(bpy.types.Operator):
    bl_idname = "mol.import_cell_pack"
    bl_label = "Load"
    bl_description = ""
    bl_options = {"REGISTER"}

    def execute(self, context):
        s = context.scene
        load_cellpack(
            file_path=s.mol_import_cell_pack_path,
            name=s.mol_import_cell_pack_name,
            node_setup=True,
        )
        return {"FINISHED"}


def panel_cellpack(layout, scene):
    layout.label(text="Load CellPack Model", icon="FILE_TICK")
    layout.separator()
    row_import = layout.row()
    row_import.prop(scene, "mol_import_cell_pack_name")
    layout.prop(scene, "mol_import_cell_pack_path")
    row_import.operator("mol.import_cell_pack")



import molecularnodes as mn


def test_coll():
    assert mn.blender.coll.mn().name == "MolecularNodes"
    assert mn.blender.coll.mn().name == "MolecularNodes"
    assert mn.blender.coll.data().name == ".MN_data"
    assert mn.blender.coll.cellpack().name == "cellpack_"
    assert mn.blender.coll.frames().name == ".data__frames"
    assert mn.blender.coll.frames("4OZS").name == ".data_4OZS_frames"



from MDAnalysis import Universe
import bpy
from bpy.props import StringProperty
from .ops import TrajectoryImportOperator
from ... import color
from ...blender import coll, nodes
import databpy
from databpy import AttributeTypes

from enum import Enum

from ..entity import EntityType

from .oxdna.OXDNAParser import OXDNAParser
from .oxdna.OXDNAReader import OXDNAReader
from .trajectory import Trajectory

DNA_SCALE = 10


class OXDNA(Trajectory):
    def __init__(self, universe: Universe, world_scale: float = 0.01):
        super().__init__(universe=universe, world_scale=world_scale * DNA_SCALE)
        self._entity_type = EntityType.MD_OXDNA
        self._att_names = (
            "base_vector",
            "base_normal",
            "velocity",
            "angular_velocity",
        )

    def _create_object(self, style: str = "oxdna", name: str = "NewUniverseObject"):
        self.object = databpy.create_object(
            name=name,
            collection=coll.mn(),
            vertices=self.univ_positions,
            edges=self.bonds,
        )
        self._update_timestep_values()

        for name in ("chain_id", "res_id", "res_name"):
            if name == "res_name":
                att_name = "res_num"
            else:
                att_name = name
            self.store_named_attribute(getattr(self, att_name), name)

        self.store_named_attribute(
            data=color.color_chains_equidistant(self.chain_id),
            name="Color",
            atype=AttributeTypes.FLOAT_COLOR,
        )

        if style:
            nodes.create_starting_node_tree(self.object, style="oxdna", color=None)

        return self.object

    def _update_positions(self, frame: int) -> None:
        super()._update_positions(frame)
        self._update_timestep_values()

    def _update_timestep_values(self):
        for name in self._att_names:
            try:
                self.store_named_attribute(
                    self.universe.trajectory.ts.data[name] * self.world_scale, name=name
                )
            except KeyError as e:
                print(e)


def load(top, traj, name="oxDNA", style="oxdna", world_scale=0.01):
    univ = Universe(top, traj, topology_format=OXDNAParser, format=OXDNAReader)
    traj = OXDNA(univ, world_scale=world_scale)
    traj.create_object(name=name, style=style)
    return traj


class MN_OT_Import_OxDNA_Trajectory(TrajectoryImportOperator):
    bl_idname = "mn.import_oxdna"

    def execute(self, context):
        load(top=self.topology, traj=self.trajectory, name=self.name)
        return {"FINISHED"}


def panel(layout, scene):
    layout.label(text="Load oxDNA File", icon="FILE_TICK")
    layout.separator()
    row = layout.row()
    row.prop(scene.mn, "import_oxdna_name")
    op = row.operator("mn.import_oxdna")
    op.name = scene.mn.import_oxdna_name
    op.topology = scene.mn.import_oxdna_topology
    op.trajectory = scene.mn.import_oxdna_trajectory
    col = layout.column(align=True)
    col.prop(scene.mn, "import_oxdna_topology")
    col.prop(scene.mn, "import_oxdna_trajectory")
